from __future__ import print_function
import numpy as np
import os
import subprocess
import astropy.table as tb 
import pickle 
import copy
from .transformation import * 
from .distribution import * 
from .popstat import *

orbdata = os.getenv('DESDATA')
if orbdata is None:
	orbdata = ''


def fibonacci_sphere(n_grid, angles = True):
	'''
	Generates a 3d Fibonacci sphere shell (see, eg, Gonzalez, A. 2010, Mathematical Geosciences, 42, 49, doi: 10.1007/s11004-009-9257-x) 
	for a certain n_grid. Generates 2n + 1 points

	Argument:
	- n_grid: grid size (integer), will generate 2n+1 points
	- angles: returns the generated angles as well as the 3D positions
	'''
	sin_phi = 2*np.arange(-n_grid, n_grid, 1, dtype=float)/(2*n_grid + 1.)
	theta = 2*np.pi * np.arange(-n_grid, n_grid, 1, dtype=float)/(0.5*(1. + np.sqrt(5)))**2
	x_shell = np.sqrt(1-sin_phi*sin_phi) * np.cos(theta)
	y_shell = np.sqrt(1-sin_phi*sin_phi) * np.sin(theta)
	z_shell = sin_phi
	if angles:
		return x_shell, y_shell, z_shell, sin_phi, theta
	else:
		return x_shell, y_shell, z_shell


class Population:
	'''
	Base class for all populations
	'''
	def __init__(self, n_objects, elementType, epoch):
		'''
			Initialization function

			Arguments:
			- n_objects: population size, will generate a (n x 6) array for the orbital elements or state vectors
			- elementType: type of element, use either 'keplerian' or 'cartesian' for Keplerian elements or state vectors, respectively
			- epoch: epoch of the elements, in years after J2000.0
		'''
		self.n_objects = n_objects
		self.elementType = elementType
		self.epoch = epoch
		self.elements = np.zeros((n_objects, 6)) 
		
		self.hasRADec = False

		if self.elementType == 'keplerian':
			self.state = 'F'
		else:
			self.state = 'T'
		self.heliocentric = False
	
	def generateMagnitudes(self, distribution, mag_type, band, colors = None, observer_pos = [1,0,0], helio = False, ecliptic = False, bands = ['g', 'r', 'i', 'z', 'Y']):
		'''
		Depends on magnitude distributions from `distribution.py`. 

		Arguments:
		- distribution: Magnitude distribution from which the magnitudes will be sampled. If a single distribution is provided, all magnitudes will be sampled from it 
		Can be a list/array of distributions, in which case the magnitude of object i will come from distribution[i]
		- mag_type: Defines if magnitudes are 'absolute' or 'apparent'. If absolute, requires observer_pos (3d array), helio and ecliptic (booleans) 
		- band: reference band for the magnitude (string)
		- colors: Dictionary of color transformations of the form band - colors[i]. colors can be None, in which case only one band is generated
		- observer_pos: 3d array or list with the location of the observer for conversion between absolute and apparent magnitudes
		- helio: boolean, uses heliocentric elements for the distance calculations (only required for orbital elements)
		- ecliptic: boolean, uses ecliptic-aligned elements for the distance calculations (only required for orbital elements)
		- bands: list of which bands the magnitudes are generated in
		'''
		if mag_type == 'absolute':
			name = 'H' 
		elif mag_type == 'apparent':
			name = 'm'
		else:
			raise ValueError("Magnitude type (mag_type) must be either absolute or apparent")
		
		if len(distribution) > 0:
			mag_band = np.zeros(self.n_objects)
			for i, dist in enumerate(distribution):
				mag_band[i] = dist.sample(1)
		else:
			mag_band = distribution.sample(self.n_objects)

		mag_table = tb.Table()

		mag_table[name + '_' + band] = 	mag_band

		for i in colors:
			mag_table[name + '_' + i] = mag_band - colors[i] 

		mag_table['ORBITID'] = range(self.n_objects)

		if mag_type == 'absolute':
			self.distanceToCenter(helio, ecliptic)
			sun_dist = self.r
			obs_dist = self.distanceToPoint(observer_pos, helio, ecliptic)

			mag_table['m_' + band] = mag_band + 5 * np.log10(sun_dist * obs_dist)

			for i in colors:
				mag_table['m_' + i] = mag_band - colors[i]  + 5 * np.log10(sun_dist * obs_dist)

		self.mag = mag_table

		band_stack = [] 

		for i in bands:
			band = mag_table['ORBITID', 'm_' + i]
			band.rename_column('m_' + i, 'MAG')
			band['BAND'] = i 
			band_stack.append(band)
		
		self.mag_obs = tb.vstack(band_stack)

	def generateLightCurve(self, lightcurve):
		'''
		Applies some sort of light curve to each detection. `lightcurve` can either be a single function in which you can call the lightcurve amplitude at a given observation time, 
		or an array containing the lightcurve for each object and indexed by ORBITID.

		Arguments:
		- lightcurve: list or individual LightCurve object
		'''

		if len(lightcurve) > 0:
			for i in range(self.n_objects):
				where = self.obs['ORBITID'] == i
				self.obs['MAG'][where] = self.obs['MAG'][where] + lightcurve[i](self.obs[where]['MJD'])
		else:
			self.obs['MAG'] = lightcurve(self.obs['MJD']) + self.obs['MAG']

	def createCopies(self, n_clones):
		'''
		Creates n_clones copies of each object, allowing for some posterior randomization
		'''

		self.elements = np.vstack([self.elements for i in range(n_clones)])
		self.n_objects *= n_clones

	def sampleElements(self, covariances, n_samples):
		'''
		Draws from the array of covariance matrices provided		

		Arguments:
		- covariances: 6D covariances of the orbital elements/state vectors
		- n_samples: number of samples to be drawn
		'''
		n_orig = copy.deepcopy(self.n_objects)
		self.covariance = covariances
		if n_samples > 1:
			self.createCopies(n_samples)

		for i in range(n_orig):
			samples = np.random.multivariate_normal(self.elements[i,:], covariances[i], n_samples)
			for j in range(n_samples):
				self.elements[i + j*n_orig,:] = samples[j]

	def distanceToCenter(self, helio = False, ecliptic = False):
		'''
		Computes distance to the center of mass of the system. helio and ecliptic are required if a transformation from orbital elements
		to cartesian vectors is needed

		Arguments:
		- helio: boolean, uses heliocentric elements for the distance calculations (only required for orbital elements)
		- ecliptic: boolean, uses ecliptic-aligned elements for the distance calculations (only required for orbital elements)
		'''

		r = dist_to_point(self.elements, self.epoch, self.elementType, np.array([0,0,0]), helio, ecliptic)
		self.r = r 

	def distanceToPoint(self, point, helio = False, ecliptic = False):
		'''
		Computes distance to the center of mass of the system. helio and ecliptic are required if a transformation from orbital elements
		to cartesian vectors is needed

		Arguments:
		- point: 3d position of the point to which the distances are being computed to
		- helio: boolean, uses heliocentric elements for the distance calculations (only required for orbital elements)
		- ecliptic: boolean, uses ecliptic-aligned elements for the distance calculations (only required for orbital elements)
		'''

		r = dist_to_point(self.elements, self.epoch, self.elementType, point, helio, ecliptic)
		return r 

	def computeStatistics(self, thresh = 90., transient_efficiency = 0.95529, drop_zero = True, linking_eff = 1):
		'''
		Computes ARC, ARCCUT and NUNIQUE for each member of the population and stores inside population.statistics. Requires a preliminary survey.observePopulation call

		Arguments:
		- thresh: threshold for triplet generation (in days). Use 60 for d < 50 au
		- transient_efficiency: efficiency of the transient generation process. Keeps transient_efficiency*100% of the transients in the catalog
		'''
		try:
			self.detections
		except:
			raise AttributeError("Population has no detections attribute. Perhaps you need to call survey.observePopulation first?")


		#consider only detections inside a CCD
		if drop_zero:
			ccdzero = np.where(self.detections['CCDNUM'] == 0)
			if len(ccdzero) > 0:
				self.detections.remove_rows(np.where(self.detections['CCDNUM'] == 0))

		#transient efficiency
		if 'RANDOM' not in self.detections.keys():
			self.detections['RANDOM'] = np.random.rand(len(self.detections))
			self.detections.remove_rows(np.where(self.detections['RANDOM'] > transient_efficiency))

		#index the table to make things easier
		self.detections.add_index('ORBITID')
		stat = tb.Table(names=['ORBITID', 'ARC', 'ARCCUT', 'NUNIQUE', 'NDETECT', 'TRIPLET', 'PAIR_1', 'PAIR_2'], dtype=['i8', 'f8', 'f8', 'i8', 'i8', 'b1', 'f8', 'f8'])
		
		ids, counts = np.unique(self.detections['ORBITID'], return_counts = True)

		for i in ids[counts > 1]:
			obj = self.detections.loc[i]
			times = np.array(obj['MJD'], dtype='float64') 
			times.sort()
			arc = np.max(times) - np.min(times)
			arccut = compute_arccut(times)
			nunique = compute_nunique(times)
			pt = find_triplet_time(times)			
			if pt[0] < thresh and pt[1] < thresh:
				has_trip = True 
			else:
				has_trip = False
			stat.add_row([i, arc, arccut, nunique, len(times), has_trip, pt[0], pt[1]])

		stack = [stat]
		ones = tb.Table()
		ones['ORBITID'] = ids[counts == 1]
		if len(ones) > 0:
			ones['ARC'] = 0.
			ones['ARCCUT'] = 0.
			ones['NUNIQUE'] = 1
			ones['NDETECT'] = 1
			ones['TRIPLET'] = False
			ones['PAIR_1'] = 0
			ones['PAIR_2'] = 0
			stack.append(ones)


		orbid = np.arange(len(self))
		zeros = tb.Table()
		zeros['ORBITID'] = orbid[np.isin(orbid, ids, invert = True)]
		if len(zeros) > 0:
			zeros['ARC'] = 0.
			zeros['ARCCUT'] = 0.
			zeros['NUNIQUE'] = 0
			zeros['NDETECT'] = 0
			zeros['TRIPLET'] = False
			zeros['PAIR_1'] = 0
			zeros['PAIR_2'] = 0
			stack.append(zeros)


		stat = tb.vstack(stack)
		stat.sort('ORBITID')

		stat['RANDOM'] = np.random.random(len(stat)) < linking_eff

		self.statistics = stat


	def randomizeElement(self, element, distribution):
		'''
		Randomizes element (must be an integer between 0 and 5) according to the provided distribution (from `distribution.py`)

		Arguments:
		- element: index of the element to be sampled
		- distribution: distribution for this element
		'''
		self.elements[:,element] = distribution.sample(self.n_objects)


	def removeMembers(self, indices):
		'''
		Deletes members of the population

		Arguments:
		- indices: array of indices that will be deletec
		'''
		self.elements = self.elements[indices]
		self.n_objects = len(self.elements)

	# Standard class methods
	def __str__(self):
		'''	
		String representation
		'''
		return "Population with {} objects. Elements are of type {}".format(self.n_objects, self.elementType)

	def __add__(self, other):
		'''
		Adds two populations together
		'''
		if self.elementType != other.elementType:
			raise ValueError("The elements of each population must be of the same type")
		if self.epoch != other.epoch:
			raise ValueError("The epochs must be the same")
		newpop = Population(self.n_objects + other.n_objects, self.elementType, self.epoch)
		newpop.elements[0:self.n_objects,:] = self.elements
		newpop.elements[self.n_objects:,:] = other.elements
		return newpop

	def __len__(self):
		return self.n_objects

	def write(self, filename):
		'''
		Saves the population to a pickle file with the specified name

		Arguments:
		- filename: filename for the saved file
		'''
		with open(filename, 'wb') as f:
			pickle.dump(self, f, protocol = 2)
	@staticmethod
	def read(filename):
		'''
		Static method, reads a file with the given filename

		Arguments:
		- filename: pickle file that will be read
		'''
		with open(filename, 'rb') as f:
			return pickle.load(f)

	def __getitem__(self, index):
		return self.elements[index,:]

	def checkInFootprint(self, footprint = orbdata + '/round17-poly.txt'):
		'''
		Checks which objects are inside the footprint
		
		Arguments:
		- footprint: file containing details of the footprint

		'''
		import matplotlib.path as mpath
		p = np.loadtxt(footprint)
		path = mpath.Path(p)

		if not self.hasRADec:
			self.getRADec()

		inside = path.contains_points(np.array([self.ra, self.dec]).T)

		self.elements = self.elements[inside]
		self.ra = self.ra[inside]
		self.dec = self.dec[inside]
		
		self.n_objects = len(self.elements)
		#self.n_grid = np.ceil(self.n_objects/2)

	def getRADec(self, heliocentric = False, ecliptic = False, observer_pos = np.array([0,0,0])):
		if self.elementType == 'keplerian':
			xv = keplerian_to_cartesian(self.elements, self.epoch, heliocentric, ecliptic)
		else:
			xv = self.elements
		xv_eq = rotate_to_ecliptic(xv)

		x_vec = xv_eq[:,:3] - observer_pos

		self.ra = np.arctan2(x_vec[:,1], x_vec[:,0]) * 180./np.pi

		self.dec = np.arcsin(x_vec[:,2]/np.sqrt(x_vec[:,0]**2 + x_vec[:,1]**2 + x_vec[:,2]**2)) * 180/np.pi

		self.hasRADec = True

	def generateSpaceRocks(self):
		import spacerocks
		import astropy.time as tm 
		import astropy.units as u 

		epoch = tm.Time(2000 + self.epoch, scale='tdb', format='jyear')
		epoch.format = 'mjd'
		units = spacerocks.Units()
		units.timescale = 'tdb'
		units.timeformat = 'mjd'
		units.speed = u.au/u.yr
		origin = 'sun' if self.heliocentric else 'ssb'

		names = [f'{i}' for i in range(len(self))]

		if self.elementType == 'keplerian':
			mu = SunGM if self.heliocentric else SolarSystemGM

			M = np.rad2deg((self.epoch - self.elements[:,5])*(np.sqrt(mu)/np.power(self.elements[:,0],3./2)))

			rock = spacerocks.SpaceRock(a = self.elements[:,0], e = self.elements[:,1], inc = self.elements[:,2],
										node = self.elements[:,3], arg = self.elements[:,4], M = M,
										epoch = len(self) * [epoch.mjd], units = units, name=names, origin=origin, frame='eclipJ2000')

		else:
			frame = 'eclipJ2000' if self.ecliptic else 'J2000'
			rock = spacerocks.SpaceRock(x = self.elements[:,0], y = self.elements[:,1], z = self.elements[:,2],
										vx = self.elements[:,3], vy = self.elements[:,4], vz = self.elements[:,5],
										epoch = len(self) * [epoch.mjd], units = units, frame = frame, name=names, origin=origin)

		return rock

class ElementPopulation(Population):
	'''
	Series of orbital elements that are submitted to `DESTracks`. The code requires (a,e,i, lan, aop, top). The user should submit at least 6 elements, and the code will convert to the right six if needed
	Units should be AU, degrees and years after J2000
	Input should be a dictionary of arrays with the proper orbital elements
	'''
	def __init__(self, elements, epoch, heliocentric = False):
		'''
		Initialization function:
		- elements: dictionary of orbital elements. Example: {'a': [30], 'e' : [0.1], 'i': [0], 'aop': [0], 'lan' : [0], 'top': [0]} 
		- epoch: epoch of the elements (years)
		- heliocentric: defines if heliocentric elements are being used


		Many different combinations of orbital elements can be provided, here is a list of all accepted elements:
		- Semi-major axis: provide as 'a' (au)
		- Eccentricity: provide as 'e'
		- Perihelion: provide as 'q' (au)
		- Inclination: provide as 'i' (degrees)
		- Argument of perihelion: provide as 'omega' or 'aop' (degrees)
		- Longitude of ascending node: provide as 'Omega' or 'lan' (degrees)
		- Longitude of perihelion: provide as 'varpi' or 'lop' (degrees)
		- Time of perihelion passage: provide as 'top' or 'T_p' (years)
		- Mean anomaly: provide as 'man' or 'M' (degrees)
		'''
		self.input = elements
		self._keys = elements.keys()
		n = len(elements[list(elements.keys())[0]])
		Population.__init__(self, n, 'keplerian', epoch)
		self._organizeElements()
		self.heliocentric = heliocentric

	def _organizeElements(self):
		mu = SunGM if self.heliocentric else SolarSystemGM
		if 'a' in self._keys:
			self.elements[:,0] = self.input['a']
		elif 'q' in self._keys and 'e' in self._keys:
			self.elements[:,0] = self.input['q']/(1 - self.input['e'])
		else:
			raise ValueError('Please provide either semi-major axis (a) or perihelion (q)/eccentricity (e) as input')
		
		if 'e' in self._keys:
			self.elements[:,1] = self.input['e']
		elif 'q' in self._keys and 'a' in self._keys:
			self.elements[:,1] = 1 - self.input['q']/self.input['a'] 
		else:
			raise ValueError("Please provide either eccentricity (e) or semi-major axis (a)/perihelion (q) as input")

		if 'i' in self._keys:
			self.elements[:,2] = self.input['i']
		else:
			raise ValueError("Please provide inclination (i) as input")

		if 'lan' in self._keys:
			self.elements[:,3] = self.input['lan']
		elif 'Omega' in self._keys:
			self.elements[:,3] = self.input['Omega']
		elif 'lop' in self._keys and 'aop' in self._keys:
			self.elements[:,3] = self.input['lop'] - self.input['aop']
		elif 'varpi' in self._keys and 'omega' in self._keys:
			self.elements[:,3] = self.input['varpi'] - self.input['omega']
		else:
			raise ValueError("Please provide either longitude of ascending node (lan) or longitude of perihelion (lop)/argument of perihelion (aop) as input")

		if 'aop' in self._keys:
			self.elements[:,4] = self.input['aop']
		elif 'omega' in self._keys:
			self.elements[:,4] = self.input['omega']
		elif 'lop' in self._keys and 'lan' in self._keys:
			self.elements[:,4] = self.input['lop'] - self.input['lan']
		elif 'varpi' in self._keys and 'Omega' in self._keys:
			self.elements[:,3] = self.input['varpi'] - self.input['Omega']
		else:
			raise ValueError("Please provide either argument of perihelion (aop) or longitude of perihelion (lop)/longitude of ascending node (lan) as input")

		if 'top' in self._keys:
			self.elements[:,5] = self.input['top']
		elif 'T_p' in self._keys:
			self.elements[:,5] = self.input['T_p']
		elif 'man' in self._keys:
			self.elements[:,5] = self.epoch - self.input['man'] * np.power(self.elements[:,0], 3./2) * (np.pi/180) / np.sqrt(mu)
		elif 'M' in self._keys:
			self.elements[:,5] = self.epoch - self.input['M'] * np.power(self.elements[:,0], 3./2) * (np.pi/180) / np.sqrt(mu)
		else:
			raise ValueError("Please provide either time of perihelion passage (top) or mean anomaly (man)/semi-major axis (a) as input")

	def randomizeAngle(self, element, min_angle = 0, max_angle = 360, uniform = True):
		'''
		Randomizes argument of perihelion 'aop' or longitude of ascending node 'lan' by either randomizing the angles or sampling all possible values
		between min_angle (= 0 deg by default) and max_angle (= 360 deg by default)

		Arguments:
		- element: 'lan' or 'aop' (or 3 or 4, for lan and aop, respectively)
		- min_angle: minimum angle (degrees)
		- max_angle: maximum angle (degrees)
		- uniform: generates in a uniform manner that covers the entire parameter space, instead of sampling from a uniform distribution
		'''
		eldic = {'lan' : 3, 'aop' : 4}
		if type(element) is str:
			element = eldic[element]
		if uniform:
			self.elements[:,element] = np.linspace(min_angle, max_angle, self.n_objects) 
		else:
			self.elements[:,element] = np.random.rand(self.n_objects)*(max_angle - min_angle) + min_angle

	def randomizeToP(self, min_angle = -180., max_angle = 180., uniform = True, heliocentric = False):
		'''
		Randomizes the Time of Perihelion passage by randomizing the true anomaly between min_angle ( =-180 deg, by default) and max_angle (180 deg)
		in either a uniform fashion (i.e. covers the entire parameter space) or drawing ran
		Arguments:
		- element: 'lan' or 'aop' (or 3 or 4, for lan and aop, respectively)
		- min_angle: minimum angle (degrees)
		- max_angle: maximum angle (degrees)
		- uniform: generates in a uniform manner that covers the entire parameter space, instead of sampling from a uniform distribution
		- heliocentric: boolean, defines the GM for the conversion between M and T_p
		'''
		mu = SunGM if heliocentric else SolarSystemGM

		if uniform:
			self.elements[:,5] = self.epoch - (np.linspace(min_angle, max_angle, self.n_objects)*np.pi/180.) * np.power(self.elements[:,0], 3./2)/np.sqrt(mu)
		else:
			self.elements[:,5] = self.epoch - (np.random.rand(self.n_objects) * (max_angle - min_angle) + min_angle) * np.pi/180. * np.power(self.elements[:,0], 3./2)/np.sqrt(mu)

	def transformElements(self, heliocentric = False, ecliptic = False):
		'''
		Transforms the population to a CartesianPopulation. Depends on being heliocentric or barycentric and ecliptic aligned or equatoriallly aligned

		Arguments:
		- heliocentric: boolean, defines if heliocentric aligned 
		- ecliptic: boolean, defines if ecliptic aligned
		'''
		xv = keplerian_to_cartesian(self.elements, self.epoch, heliocentric, ecliptic)
		xv_dict = {'x' : xv[:,0], 'y' : xv[:,1], 'z' : xv[:,2], 'vx' : xv[:,3], 'vy' : xv[:,4], 'vz' : xv[:,5]}
		return CartesianPopulation(xv_dict, self.epoch)

	def toBarycentric(self, barycenter_coordinates):
		'''
		Transforms a set of heliocentric orbital elements to barycentric orbital elements
		Arguments:
		- barycenter_coordinates: 6D barycenter state vector (AU, AU/yr)
		'''
		if not self.heliocentric:
			raise ValueError("Population is already in barycentric elements")
		else:
			new_aei = helio_to_bary(self.elements, "keplerian", self.epoch, barycenter_coordinates)
			aei_dict = {'a' : new_aei[:,0], 'e' : new_aei[:,1], 'i' : new_aei[:,2], 'lan' : new_aei[:,3], 'aop' : new_aei[:,4], 'T_p' : new_aei[:,5]}
			return ElementPopulation(aei_dict, self.epoch, False)




class CartesianPopulation(Population):
	'''
	Base class for a population using state vectors to be submitted to `DESTracks`. Positions are always in AU, velocities in AU/yr
	Input: a dictionary with (x,y,z,vx,vy,vz) entries 
	'''
	def __init__(self, elements, epoch, ecliptic = False, heliocentric = False):
		'''
		Initialization function

		Arguments:
		- dictionary of elements containing 'x', 'y', 'z', 'vx', 'vy', 'vz'
		- epoch: epoch of the elements (years)
		- heliocentric: boolean, defines if heliocentric elements are being used
		- ecliptic: boolean, defines if ecliptic-aligned or equatorially aligned
		'''
		self.input = elements
		n = len(elements[list(elements.keys())[0]])
		Population.__init__(self, n, 'cartesian', epoch)
		self.elements[:,0] = self.input['x']
		self.elements[:,1] = self.input['y']
		self.elements[:,2] = self.input['z']
		self.elements[:,3] = self.input['vx']
		self.elements[:,4] = self.input['vy']
		self.elements[:,5] = self.input['vz']

		self.ecliptic = ecliptic
		self.heliocentric = heliocentric

	def transformElements(self, heliocentric = False, ecliptic = False):
		'''
		Transforms the population to a ElementPopulation. Depends on being heliocentric or barycentric and ecliptic aligned or equatoriallly aligned
		
		Arguments:
		- heliocentric: boolean, defines if heliocentric elements are being used
		- ecliptic: boolean, defines if ecliptic-aligned or equatorially aligned
		'''
		aei = cartesian_to_keplerian(self.elements, self.epoch, heliocentric, ecliptic)
		aei_dict = {'a' : aei[:,0], 'e' : aei[:,1], 'i' : aei[:,2], 'Omega' : aei[:,3], 'omega' : aei[:,4], 'T_p' : aei[:,5]}
		return ElementPopulation(aei_dict, self.epoch)

	def rotateElements(self):
		'''
		Rotates between equatorial and ecliptic state vectors
		'''
		self.elements = rotate_to_ecliptic(self.elements, not self.ecliptic)
		self.ecliptic = not self.ecliptic
	

class IsotropicPopulation(CartesianPopulation):
	'''
	Generates two Fibonacci spheres (one for velocities and one for positions) objects are sampled across all possible parameters
	See detailed description in Bernardinelli et al 2020a for the detailed explanation
	'''
	def __init__(self, n_grid, epoch, drop_outside = True, ecliptic = False, footprint = orbdata + '/round17-poly.txt'):
		'''
		Initialization function

		Arguments:
		- n_grid: grid size for the Fibonacci sphere, creates 2n+1 points
		- epoch: epoch for the elements (years)
		- drop_outside: boolean, deletes objects outside the specified footprint (eg the DES footprint)
		- ecliptic: boolean, defines if ecliptic aligned
		- footprint: file containing details of the footprint
		'''
		### Warning: n_grid defines the _grid size_, which leads to 2n+1 objects
		size = len(np.arange(-n_grid, n_grid, 1, dtype=float))
		Population.__init__(self, size, 'cartesian', epoch)
		self.n_grid = n_grid
		self._generateShell()
		self.hasRADec = True
		if drop_outside:
			self.checkInFootprint(footprint)
			self.n_grid = int(self.n_objects/2)
		self._generateVelocityShell()
		
		self.ecliptic = ecliptic

		if ecliptic:
			print("DESTracks requires equatorial state vectors! Remember to rotate first.")

	def _generateShell(self):
		'''
		Places objects on a spherical shell of radius 1 AU
		'''
		x,y,z, sin_dec, ra = fibonacci_sphere(self.n_grid)
		self.elements[:,0] = x
		self.elements[:,1] = y
		self.elements[:,2] = z

		self.ra = np.mod(180*ra/np.pi, 360)
		self.ra[self.ra>180] -= 360 

		self.dec = 180*np.arcsin(sin_dec)/np.pi


	def generateDistances(self, distribution):
		'''
		Generates distances according to the provided distribution

		Arguments:
		- distribution: distribution object of distances to be used
		''' 
		r = distribution.sample(self.n_objects)

		self.elements[:,0] *= r
		self.elements[:,1] *= r
		self.elements[:,2] *= r
		self.r = r

	def _generateVelocityShell(self):
		vx, vy, vz = fibonacci_sphere(self.n_grid, False)
		perm = np.random.permutation(self.n_objects)
		self.elements[:,3] = vx[perm]
		self.elements[:,4] = vy[perm]
		self.elements[:,5] = vz[perm]

	def generateVelocities(self, distribution):
		'''
		Places the velocity from the 1 AU/yr shell into more realistic units scaled by the escape velocity and the provided distribution

		Arguments:
		- distribution: distribution object of distances to be used

		'''
		v_escape = np.sqrt(2*SolarSystemGM/self.r)
		v_scale = distribution.sample(self.n_objects)

		self.elements[:,3] *= v_escape * v_scale
		self.elements[:,4] *= v_escape * v_scale
		self.elements[:,5] *= v_escape * v_scale

		self.v = v_escape * v_scale




from __future__ import print_function
import numpy as np
from itertools import chain
from .ccd import *
import os 
import subprocess

from scipy.spatial import cKDTree

import astropy.table as tb 
from astropy.time import Time  
from astropy import units as u

import spacerocks 
from spacerocks.units import Units
from spacerocks.simulation import Simulation
from spacerocks.model import PerturberModel, builtin_models
from spacerocks.cbindings import correct_for_ltt_destnosim
from spacerocks.observer import Observer
from spacerocks.constants import epsilon

class DECamExposure:
	'''
	Base class for a DECam exposure. Contains a bunch of useful functions to deal with coordinates
	near the exposure
	'''
	def __init__(self, expnum, ra, dec, mjd_mid, band):
		'''
		Initialization function

		Arguments:
		- expnum: Exposure number 
		- ra: R.A. pointing (degrees)
		- dec: Declination pointing (degrees)
		- mjd_mid: midpoint time of the exposure (mjd)
		- band: exposure filter
		'''
		self.expnum = expnum
		self.ra = ra
		if self.ra > 180:
			self.ra -= 360
		self.dec = dec
		self.mjd = mjd_mid
		self.band = band
		self.corners = {}

	def gnomonicProjection(self, ra, dec):
		'''
		Computes the Gnomonic projection of the positions supplied centered on this exposure
		
		Arguments:
		- ra: list (or numpy array) of R.A.s (degrees)
		- dec: list (or numpy array) of Decs (degrees)
		
		Returns Gnomonic x and y, in degrees
		'''
		ra_list = np.array(ra)
		dec_list = np.array(dec)
		ra_list[np.where(ra_list > 180)] = ra_list[np.where(ra_list > 180)] - 360
	
	
		c_dec_0 = np.cos(np.pi*self.dec/180.)
		s_dec_0 = np.sin(np.pi*self.dec/180.)
		c_dec   = np.cos(np.pi*dec_list/180.)
		s_dec   = np.sin(np.pi*dec_list/180.)
		c_ra    = np.cos(np.pi*(ra_list-self.ra)/180.)
		s_ra    = np.sin(np.pi*(ra_list-self.ra)/180.)

		cos_c = s_dec_0*s_dec + c_dec_0*c_dec*c_ra

		x = c_dec*s_ra/cos_c
		y = (c_dec_0*s_dec - s_dec_0*c_dec*c_ra)/cos_c

		return 180*x/np.pi, 180*y/np.pi

	def inverseGnomonic(self, x, y):
		'''
		Inverts Gnomonic positions from a projection centered on the exposure
		
		Arguments:
		- x: list or array of Gnomonic x (degrees)
		- y: list or array of Gnomonic y (degrees)

		Returns R.A. and Decs, in degrees
		'''
		x = np.array(x) * np.pi/180.
		y = np.array(y) * np.pi/180.
		ra_rad = np.pi*self.ra/180.
		dec_rad = np.pi*self.dec/180.
		den = np.sqrt(1 + x**2 + y**2)
		sin_dec = (np.sin(dec_rad) + y*np.cos(dec_rad))/den
		sin_ra = x/(np.cos(dec_rad) * den)
		return 180.*(np.arcsin(sin_ra))/np.pi + self.ra, 180.*np.arcsin(sin_dec)/np.pi 
 

	def checkInCCDFast(self, ra_list, dec_list, ccd_tree = None, ccd_keys = None, ccdsize = 0.149931):
		'''
		Checks if a list of RAs and Decs are inside a DECam CCD in an approximate way, returns indices of positions that are inside a CCD and which CCD they belong to

		Arguments:
		- ra_list: list of R.A.s (degrees)
		- dec_list: list of Decs (degrees)
		- ccd_tree: kD tree of the CCD positions. Can be generated using ccd.create_ccdtree()
		- ccd_keys: list of CCD correspondences with the kD tree. Can be generated using ccd.create_ccdtree()
		- ccdsize: Size, in degrees, of the Gnomonic x direction (smallest side).

		Returns:
		- List of CCD positions for each RA/Dec pair (empty if there is no CCD match)
		- List of CCDs in which each point belongs to
		'''
		if ccd_tree == None:
			ccd_tree, ccd_keys = create_ccdtree()

		x, y = self.gnomonicProjection(ra_list, dec_list)

		tree_data = np.array([x, 2*y]).T
		index = np.arange(len(ra_list))
		if len(x) > 0:
			tree = cKDTree(tree_data)
			## CCD size = 0.149931 deg
			inside_CCD = ccd_tree.query_ball_tree(tree, ccdsize, p = np.inf)
			#this is probably the most complicated Python line ever written
			if inside_CCD != None: 
				ccd_id = [len(inside_CCD[i])*[ccdnums[ccd_keys[i]]] for i in range(len(inside_CCD)) if len(inside_CCD[i]) > 0]
				inside_CCD = np.array(list(chain(*inside_CCD)))
				if len(inside_CCD) > 0:
					return index[inside_CCD], list(chain(*ccd_id))
				else:
					return [], None
			else:
				return [], None
		else:
			return [], None

	def __str__(self):
		'''
		String representation function
		'''
		return 'DECam exposure {} taken with {} band. RA: {} Dec: {} MJD: {}'.format(self.expnum, self.band, self.ra, self.dec, self.mjd)

	def checkInCCDRigorous(self, ra_list, dec_list, ccd_list):
		'''
		Checks if the object is really inside the CCD using the corners table and a ray tracing algorithm
		
		Arguments:
		- ra_list: list of R.A.s (degrees)
		- dec_list: list of Decs (degrees)
		- ccd_list: list of CCDs to be checked by the ray tracing algorithm (coming, eg, from self.checkInCCDFast). Must match RA/Dec shape 

		Returns:
		- List of booleans if the RA/Dec pair belongs to its correspondent CCD.
		'''

		try:
			self.corners
		except:
			raise AttributeError("Exposure doesn't have dict of corners. Perhaps run Survey.collectCorners?")
		
		inside = []

		for ra, dec, ccd in zip(ra_list, dec_list, ccd_list):
			try:
				inside.append(ray_tracing(ra, dec, self.corners[ccd]))
			except KeyError:
				inside.append(False)
		
		return inside




class Survey:
	'''
	A survey is just a series of exposures. Ideally, we'd have one `exposure.positions.fits` file for DESTracks usage
	This function includes convenience calls for $ORBITSPP to generate positions of a population of objects
	'''
	def __init__(self, expnum, ra, dec, mjd, band, track = None, corners = None):
		'''
		Initialization function

		Arguments:
		- expnum: list of exposure numbers
		- ra: list of R.A. pointings
		- dec: list of Dec pointings
		- mjd: list of midpoint MJDs for the exposures
		- band: list of filters for each exposure
		- track: Path of FITS file containing the exposures for use with $ORBITSPP
		- corners: Path of CCD corners FITS file containing the CCD corners of all exposures for use with $ORBITSPP
		'''
		self.ra = ra
		self.ra[self.ra > 180] -= 360 
		self.dec = dec 
		self.mjd = mjd 
		self.expnum = expnum
		self.band = band

		self.track = track
		self.corners = corners

		self._hascorners = False

	def createExposures(self):
		'''
		Creates a dictionary of DECamExposures for the Survey inside self.exposures
		'''
		self.exposures = {}
		for ra,dec,mjd,n,b in zip(self.ra, self.dec, self.mjd, self.expnum, self.band):
			self.exposures[n] = DECamExposure(n, ra, dec, mjd, b)

	def _createEarthSpaceRock(self):
		'''
		Creates the pre-computed Earth positions from SpaceRocks
		'''
		import astropy.units as u
		table = tb.Table.read(self.track)

		self.times = Time(self.mjd, format='mjd', scale='tdb')
		units = spacerocks.Units()
		units.timescale = 'tdb'
		units.timeformat = 'mjd'
		units.speed = u.au/u.yr

		self.earth =  spacerocks.SpaceRock(x = table['observatory'][:,0], y = table['observatory'][:,1], z = table['observatory'][:,2],
										   vx = table['velocity'][:,0], vy = table['velocity'][:,1], vz = table['velocity'][:,2],
										   epoch = self.times, units = units, origin = 'ssb', frame='J2000')

	def createObservations(self, population, outputfile, useold = False, ra0 = 10, dec0 = -20, radius = 85):
		'''
		Calls $ORBITSPP/DESTracks to generate observations for the input population, saves them in the outputfile 
		and returns this table

		Arguments:
		- population: Population object from tno/population containing the input orbits
		- outputfile: Path for the output FITS file where the observations will be saved, .fits extension will be appended
		- useold: boolean, if True will check if outputfile already exists and read it, skipping the DESTracks call
		- ra0: R.A. center of the observation field
		- dec0: Dec center of the observation
		- radius: search radius for the exposures
		
		For DES usage, the last three parameters should remain constant!

		Results are stored in population.observations
		'''
		if useold and os.path.exists(outputfile + '.fits'):
			population.observations =  tb.Table.read(outputfile + '.fits')
			return None

		if population.heliocentric:
			raise ValueError("Please use barycentric elements!")

		orbitspp = os.getenv('ORBITSPP')
		with open('{}.txt'.format(outputfile), 'w') as f:
			for j,i in enumerate(population.elements):
				print(j, i[0],i[1],i[2],i[3],i[4],i[5], file = f)
				
		with open('{}.txt'.format(outputfile), 'r') as f:

			print(' '.join([orbitspp + '/DESTracks', '-cornerFile={}'.format(self.corners), 
							'-exposureFile={}'.format(self.track), '-tdb0={}'.format(population.epoch), '-positionFile={}.fits'.format(outputfile)
							,'-readState={}'.format(population.state) , '-ra0={}'.format(ra0), 
							'-dec0={}'.format(dec0),'-radius={}'.format(radius), '< {}.txt'.format(outputfile)]))

			subprocess.run(' '.join([orbitspp + '/DESTracks', '-cornerFile={}'.format(self.corners), 
							'-exposureFile={}'.format(self.track), '-tdb0={}'.format(population.epoch), '-positionFile={}.fits'.format(outputfile)
							,'-readState={}'.format(population.state) , '-ra0={}'.format(ra0), 
							'-dec0={}'.format(dec0),'-radius={}'.format(radius), ' < {}.txt'.format(outputfile)]), stdin = f, shell = True)

		if not os.path.exists(outputfile + '.fits'):
			raise ValueError("$ORBITSPP call did not terminate succesfully!")


		population.observations =  tb.Table.read(outputfile + '.fits')

	def createObservationsSpacerocks(self, population, progress=True):
		'''
		Calls the Spacerocks backend to generate observations for the input population


		Arguments:
		- population: Population object for the input orbits
		'''
		## first set up times and do spacerock stuff

		#self.createEarthSpaceRock()
		self.times = Time(self.mjd, format='mjd', scale='utc')
		rocks = population.generateSpaceRocks()

		units = Units()
		units.timescale = 'utc'
		units.timeformat = 'jd'
		units.mass = u.M_sun
        
		spiceids, kernel, masses = builtin_models['ORBITSPP']
		model = PerturberModel(spiceids=spiceids, masses=masses)
		
		sim = Simulation(model=model, epoch=self.times.jd[0], units=units)
		sim.add_spacerocks(rocks)
		sim.integrator = 'leapfrog'

		ras = np.array([])
		decs = np.array([])
		orbitid = np.array([])
		expnum = np.array([]) 
		oidlist = np.arange(len(population))
		if progress == True:
			from rich.progress import track
			epochs = track(range(len(self.times.jd)))
		else:
			epochs = range(len(self.times.jd))

		for i in epochs:
			sim.integrate(self.times.jd[i], exact_finish_time=1)
			a = np.zeros((sim.N, 3), dtype=np.double)
			b = np.zeros((sim.N, 3), dtype=np.double)
			sim.serialize_particle_data(xyz=a, vxvyvz=b)
			x, y, z = a.T
			vx, vy, vz = b.T
        
			x = np.ascontiguousarray(x)[sim.N_active:]
			y = np.ascontiguousarray(y)[sim.N_active:]
			z = np.ascontiguousarray(z)[sim.N_active:]
			vx = np.ascontiguousarray(vx)[sim.N_active:]
			vy = np.ascontiguousarray(vy)[sim.N_active:]
			vz = np.ascontiguousarray(vz)[sim.N_active:]

			observer = Observer(epoch=self.times.jd[i], obscode='W84', units=units)
			ox = observer.x.au.astype(np.double)
			oy = observer.y.au.astype(np.double)
			oz = observer.z.au.astype(np.double)
			ovx = observer.vx.value.astype(np.double)
			ovy = observer.vy.value.astype(np.double)
			ovz = observer.vz.value.astype(np.double)
			
			# Compute ltt-corrected topocentroc Ecliptic coordinates
			xt, yt, zt = correct_for_ltt_destnosim(x, y, z, vx, vy, vz, ox, oy, oz, ovx, ovy, ovz)
			lon = np.arctan2(yt, xt)
			lat = np.arcsin(zt / np.sqrt(xt**2 + yt**2 + zt**2))
			dec = np.degrees(np.arcsin(np.sin(lat) * np.cos(epsilon) + np.cos(lat) * np.sin(lon) * np.sin(epsilon)))
			ra = np.degrees(np.arctan2((np.cos(lat) * np.cos(epsilon) * np.sin(lon) - np.sin(lat) * np.sin(epsilon)), np.cos(lon) * np.cos(lat)))

			ra[ra>180] -= 360

			dist_center = np.sqrt( ((ra - self.ra[i]) * np.cos(self.dec[i] * np.pi/180))**2 + (dec - self.dec[i])**2)

			keep = dist_center < 1.1

			ras = np.append(ras, ra[keep])
			decs = np.append(decs, dec[keep])
			orbitid = np.append(orbitid, oidlist[keep])
			expnum = np.append(expnum, len(oidlist[keep]) * [self.expnum[i]])


		del x, y, z, vx, vy, vz, a, b, sim, xt, yt, zt, ox, oy, oz, observer
		## gather data into something useable
		t = tb.Table()
		t['RA'] = ras
		del ras
		#t['RA'][t['RA'] > 180] -= 360
		t['DEC'] = decs
		del decs
		t['EXPNUM'] = expnum
		t['EXPNUM'] = t['EXPNUM'].astype('int32')
		del expnum
		t['ORBITID'] = orbitid
		t['ORBITID'] = t['ORBITID'].astype('int64')
		del orbitid

		exp = tb.Table()
		exp['EXPNUM'] = np.array(self.expnum)
		exp['RA_CENTER'] = np.array(self.ra)
		#exp['RA_CENTER'][exp['RA_CENTER'] > 180] -= 360
		exp['DEC_CENTER'] = np.array(self.dec) 
		#return t, exp
		t = tb.join(t, exp)

		
		x, y = bulk_gnomonic(np.array(t['RA']), np.array(t['DEC']), np.array(t['RA_CENTER']), np.array(t['DEC_CENTER']))
		#rescale for kD tree
		theta = np.array([x, 2*y]).T
		del t['RA_CENTER', 'DEC_CENTER']

		ccd_tree, ccd_keys = create_ccdtree()

		tree = cKDTree(theta)
		# kD tree ccd checker
		inside_ccd = ccd_tree.query_ball_tree(tree, 0.149931 * 1.01, p = np.inf)
		if inside_ccd != None: 
			ccd_id = [len(inside_ccd[i])*[ccdnums[ccd_keys[i]]] for i in range(len(inside_ccd)) if len(inside_ccd[i]) > 0]
			inside_ccd = np.array(list(chain(*inside_ccd)))
			if len(inside_ccd) > 0:
				ccdlist = list(chain(*ccd_id))
			else:
				print('No observations!')
				self.population = tb.Table(column=['RA', 'DEC', 'EXPNUM', 'ORBITID'])
				return None
		else:
			print('No observations!')
			self.population = tb.Table(column=['RA', 'DEC', 'EXPNUM', 'ORBITID'])
			return None 

		outside_ccd = np.arange(len(t))
		outside_ccd = outside_ccd[np.isin(outside_ccd, inside_ccd, invert = True)]
		outccd = t[outside_ccd]
		t = t[inside_ccd]
		t['CCDNUM'] = ccdlist

		# we need the exposure objects here for the proper CCD stuff

		try:
			self.exposures
		except AttributeError:
			self.createExposures()

		if not self._hascorners:
			self.collectCorners()


		inside_ccd = np.zeros(len(t), dtype='bool')

		for i in range(len(t)):

			# proper corners
			if t[i]['CCDNUM'] not in self.exposures[t[i]['EXPNUM']].corners:
				ins = False
			else:
				ins = ray_tracing(t[i]['RA'], t[i]['DEC'], self.exposures[t[i]['EXPNUM']].corners[t[i]['CCDNUM']])

			inside_ccd[i] = ins 

		outccd = tb.vstack([outccd, t[~inside_ccd]])
		outccd['CCDNUM'] = 0
		outccd = tb.unique(outccd)

		obs = t[inside_ccd]
		obs = tb.vstack([t, outccd])
		obs.sort(['ORBITID','EXPNUM'])

		population.observations = obs['RA', 'DEC', 'EXPNUM', 'CCDNUM', 'ORBITID']
	

	def __getitem__(self, key):
		'''
		Allows DECamExposures to be accessed by indexing the Survey object
		'''
		try:
			return self.exposures[key]
		except AttributeError:
			raise AttributeError("Survey does not have a list of DECamExposures!")
		except KeyError:
			raise KeyError("Exposure {} not in survey".format(key))

	def __len__(self):
		'''
		Returns the number of exposures
		'''
		return len(self.expnum)

	def collectCorners(self):
		'''	
		Uses the CCD corners table to build a list of CCDs
		'''
		if self.corners == None:
			raise ValueError("No table of CCD corners! Set Survey.corners first.")
		else:
			corners = tb.Table.read(self.corners)
		try:
			self.exposures
		except AttributeError:
			self.createExposures()

		corners = corners[np.isin(corners['expnum'], self.expnum)]
		#corners.add_index('expnum')

		for i in corners:
			rac = i['ra'][:-1]
			rac[rac>180] -= 360 
			decc = i['dec'][:-1]

			self.exposures[i['expnum']].corners[i['ccdnum']] = np.array([rac,decc]).T

		self._hascorners = True








'''
Code adapted from https://tmramalho.github.io/blog/2013/12/16/how-to-do-inverse-transformation-sampling-in-scipy-and-numpy/
'''

import numpy as np
import scipy.interpolate as interpolate

def inverse_transform_sampling(x_values, y_values, n_bins=40, n_samples=1000, kind = 'cubic'):
    '''
    Performs an inverse transform sampling algorithm for a distribution with provided x and y values

    Arguments:
    - x_values: range of the distribution
    - y_values: correspondend y values for the pdf, must be normalized
    - n_bins: number of bins for histogramming
    - n_samples: number of samples for the histogramming process
    - kind: interpolation type from scipy.interpolate.interp1d
    '''
    hist, bin_edges = np.histogram(x_values, bins=n_bins, density=True, weights = y_values)
    cum_values = np.zeros(bin_edges.shape)
    cum_values[1:] = np.cumsum(hist*np.diff(bin_edges))
    inv_cdf = interpolate.interp1d(cum_values, bin_edges, kind)
    r = np.random.rand(n_samples)
    return inv_cdf(r)


def inverse_cdf(x_values, y_values, n_bins=40, kind = 'cubic'):
    '''
    Performs an inverse transform sampling algorithm for a cdf with provided x and y values

    Arguments:
    - x_values: range of the distribution
    - y_values: correspondend y values for the cdf, must be normalized
    - n_bins: number of bins for histogramming
    - n_samples: number of samples for the histogramming process
    - kind: interpolation type from scipy.interpolate.interp1d
    '''

    hist, bin_edges = np.histogram(x_values, bins=n_bins, density=True, weights = y_values)
    cum_values = np.zeros(bin_edges.shape)
    cum_values[1:] = np.cumsum(hist*np.diff(bin_edges))
    inv_cdf = interpolate.interp1d(cum_values, bin_edges, kind)
    return inv_cdf

def inverse_cdf_histogram(hist, bin_edges, kind = 'cubic'):
    '''
    Assumes that the histogram already exists somewhere, returns the sampling cdf

    Arguments:
    - hist: histogram values at each bin (numpy style) 
    - bin_edges: edges of the bins (numpy style)
    - kind: interpolation type from scipy.interpolate.interp1d
    '''
    cum_values = np.zeros(bin_edges.shape)
    cum_values[1:] = np.cumsum(hist*np.diff(bin_edges))
    inv_cdf = interpolate.interp1d(cum_values, bin_edges, kind)
    return inv_cdf


from __future__ import print_function
import numpy as np 
from numba import vectorize

'''
Solar system constants
Information from https://github.com/gbernstein/gbutil/blob/master/include/AstronomicalConstants.h
'''
# ----------------------------------------------------------------------------------------------------------------
SunGM 		=	 4.*np.pi*np.pi/1.000037773533  #solar gravitation
MercuryGM 	= 	6.55371264e-06
VenusGM 	=   9.66331433e-05
EarthMoonGM =	1.20026937e-04
MarsGM 		=   1.27397978e-05;
JupiterGM   =	3.76844407e-02 + 7.80e-6
SaturnGM 	=  1.12830982e-02 + 2.79e-6
UranusGM 	=  1.72348553e-03 + 0.18e-6
NeptuneGM 	= 2.03318556e-03 + 0.43e-6

SolarSystemGM = SunGM + MercuryGM + VenusGM + EarthMoonGM + MarsGM + JupiterGM + SaturnGM + UranusGM + NeptuneGM

EclipticInclination = 23.43928 * np.pi/180
# ----------------------------------------------------------------------------------------------------------------


def R(omega, Omega, i):
	'''
	Computes the rotation matrix :math:`R = R(\\Omega, \\omega, i)` that maps from the plane of the ellipse to 3D space aligned with the ecliptic plane
	Args:
		omega (float, can be a n-dimensional array): argument of perihelion (in degrees)
		Omega (float, can be a n-dimensional array): longitude of ascending node (in degrees)
		i (float, can be a n-dimensional array): inclination (in degrees)
	Returns:
		3x3 array with the rotation matrix, if the elements are an n-dimensional array, the shape is (n,3,3)

	'''


	cO = np.cos(np.pi*Omega/180)
	sO = np.sin(np.pi*Omega/180)
	co = np.cos(np.pi*omega/180)
	so = np.sin(np.pi*omega/180)
	ci = np.cos(np.pi*i/180)
	si = np.sin(np.pi*i/180)

	R = np.array([[cO * co - sO * so * ci, - cO * so - sO * co * ci, sO * si],
				[sO * co + cO * so * ci, -sO * so + cO * co * ci, -cO * si],
				[si * so, si*co, ci]])


	if np.isscalar(omega):
		return R
	else:
		return np.transpose(R,(2,0,1))

def cartesian_to_keplerian(cartesian, epoch, helio = False, ecliptic = False):
	'''
	Goes from ecliptic or equatorially oriented state vector (i.e. cartesian representation) to orbital elements.

	Args:
		cartesian (n x 6 array): array with the (x,y,z,vx,vy,vz) Cartesian elements
		epoch (float): Epoch at which the elements are defined, in years
		helio (bool, default = False): Switches between barycentric and heliocentric elements (False is barycentric)
		ecliptic (bool, default = False): Use true if input elements are equatorially-aligned (ICRS)
	Returns:
		aei (n x 6 array): array with (a,e,i,Omega,omega,T_p) elements
	'''
	mu = SunGM if helio else SolarSystemGM

	xv = np.zeros_like(cartesian)
	if not ecliptic:
		cosEcl = np.cos(EclipticInclination)
		sinEcl = np.sin(EclipticInclination)
		xv[:,0] = cartesian[:,0]
		xv[:,3] = cartesian[:,3]
		xv[:,1], xv[:,2] = cosEcl * cartesian[:,1] + sinEcl * cartesian[:,2], -sinEcl * cartesian[:,1] + cosEcl * cartesian[:,2]
		xv[:,4], xv[:,5] = cosEcl * cartesian[:,4] + sinEcl * cartesian[:,5], -sinEcl * cartesian[:,4] + cosEcl * cartesian[:,5]
	else:
		xv = cartesian


	x = np.sqrt(xv[:,0]*xv[:,0] + xv[:,1]*xv[:,1]+ xv[:,2]*xv[:,2])
	vsq_mu = (xv[:,3]**2 + xv[:,4]**2 + xv[:,5]**2)/mu

	inv_a = 2./x - vsq_mu
	a = 1./inv_a

	x_dot_v = (xv[:,0] * xv[:,3] + xv[:,1] * xv[:,4] + xv[:,2] * xv[:,5])
	pref = (vsq_mu - 1./x)
	e_vec = (pref * xv[:,0:3].T -  x_dot_v * xv[:,3:6].T/mu).T
	e = np.sqrt(e_vec[:,0] * e_vec[:,0] + e_vec[:,1] * e_vec[:,1] + e_vec[:,2] * e_vec[:,2])
	h_vec = np.cross(xv[:,0:3], xv[:,3:6])
	n_vec = np.cross(np.array([0,0,1]), h_vec)
	h = np.sqrt(h_vec[:,0] * h_vec[:,0] + h_vec[:,1] * h_vec[:,1] + h_vec[:,2] * h_vec[:,2])
	n = np.sqrt(n_vec[:,0] * n_vec[:,0] + n_vec[:,1] * n_vec[:,1] + n_vec[:,2] * n_vec[:,2])
	cos_i = h_vec[:,2]/h 

	cosOmega = n_vec[:,0]/n 
	cosomega = (n_vec[:,0] * e_vec[:,0] + n_vec[:,1] * e_vec[:,1] + n_vec[:,2] * e_vec[:,2])/(n*e)

	i = np.arccos(cos_i) * 180./np.pi
	Omega = np.arccos(cosOmega) * 180./np.pi
	omega = np.arccos(cosomega) * 180./np.pi
	Omega[np.where(n_vec[:,1] < 0)] = 360 - Omega[np.where(n_vec[:,1] < 0)]
	omega[np.where(e_vec[:,2] < 0)] = 360 - omega[np.where(e_vec[:,2] < 0)]

	p = h*h/mu
	b = a * np.sqrt(1 - e*e)

	xbar = (p - x)/e 
	ybar = x_dot_v/e * np.sqrt(p/mu)

	E = np.arctan2(ybar/b, xbar/a + e)

	M = E - e*np.sin(E)

	T_p = epoch - M * np.sqrt(a**3/mu)

	aei = np.zeros_like(xv)

	aei[:,0] = a
	aei[:,1] = e 
	aei[:,2] = i 
	aei[:,3] = Omega
	aei[:,4] = omega
	aei[:,5] = T_p

	return aei

def q(E, a, e):
        '''
        Computes

        :math:`\\mathbf{q} = \(a(\\cos(E) - e), a\\sqrt{1-e^2}\\sin(E),0 \)`,

        the coordinate vector on the plane of the ellipse

        Args:
        	E (float, array): eccentric anomaly
        	a (float, array): semi-major axis
        	e (float, array): eccentricity
        Returns:
        	q (n x 3 array): coordinate vector on the plane of the ellipse
        '''     
        q1 = a*(np.cos(E)-e)
        q2 = a*np.sqrt(1-e**2)*np.sin(E)

        if np.isscalar(a):
        	return np.array([q1,q2,0])
        else:
        	return np.array([q1,q2,np.zeros_like(a)])

def dqdt(E, a, e, mu):
        '''
        Computes

        :math:`\\frac{\\mathrm{d}\\mathbf{q}}{\\mathrm{d} t} = \(- \\frac{n a \\sin E}{1 - e \\cos E}, \\frac{n a \\sqrt{1-e^2} \\cos E}{1 - e \\cos E},0 \)`,

        the velocity on the plane of the ellipse, where :math:`n = \\sqrt{G (M + m)}a^{-3\2} \\aprox \\sqrt{GM}a^{-3/2} = 2 \\pi a^{-3/2}` 
        in our units and heliocentric coordinates.
        
        Args:
        	E (float, array): eccentric anomaly
        	a (float, array): semi-major axis
        	e (float, array): eccentricity
        	mu (float): Reduced mass
        Returns:
        	dqdt (n x 3 array): velocity vector on the plane of the ellipse
        '''     

             
        n = np.sqrt(mu)/np.power(a,3./2)
        den = 1 - e*np.cos(E)
        q1 = -n*a*np.sin(E)/den
        q2 = n*a*np.sqrt(1-e**2)*np.cos(E)/den
        if np.isscalar(a):
        	return np.array([q1,q2,0])
        else:
        	return np.array([q1,q2,np.zeros_like(a)])


@vectorize(["float32(float32,float32)", "float64(float64,float64)"],nopython=True)
def solve_anomaly(e, M0):
	'''
	Super fast way of iteratively computing the eccentric anomaly given the mean anomaly. 
	Uses numba vectorization

	Args:
		e (float, array): eccentricity
		M0 (float, array): Mean anomaly
	Returns:
		E (float, array): Eccentric anomaly at input epoch (i.e. in M0)
	'''
	sol = M0
	delta = 0.0
	ones = 1.0
	for i in range(1000):
		delta = (M0 - (sol - e * np.sin(sol)))/(ones - e*np.cos(sol))
		sol += delta
	return sol

def rotate_to_ecliptic(xv, inverse = False):
	'''
	Rotates a 6D element vector between the ecliptic and equatorial frames
	Args:
		xv (n x 6 array): Cartesian representation in ecliptic or equatorial coordinates
		inverse (bool, default = False): triggers between ecliptic or equatorial transformations
	Returns:
		xv (n x 6 array): rotated array
	'''
	cosEcl = np.cos(EclipticInclination)
	sinEcl = -np.sin(EclipticInclination)
	if inverse:
		sinEcl = - sinEcl
	xv[:,1], xv[:,2] = cosEcl * xv[:,1] + sinEcl * xv[:,2], -sinEcl * xv[:,1] + cosEcl * xv[:,2]
	xv[:,4], xv[:,5] = cosEcl * xv[:,4] + sinEcl * xv[:,5], -sinEcl * xv[:,4] + cosEcl * xv[:,5]

	return xv


def keplerian_to_cartesian(keplerian, epoch, helio = False, ecliptic = True):
	'''
	Goes from Keplerian elements to ecliptic or equatorial state vectors
	See chapter 1 of Modern Celestial Mechanics - Morbidelli for example

	Args:
		keplerian (n x 6 array): Keplerian orbital elements, (a,e,i,omega,Omega,T_p)
		epoch (float): Epoch at which the elements are defined, in years
		helio (bool, default = False): Switches between barycentric and heliocentric elements (False is barycentric)
		ecliptic (bool, default = True): Use true if output elements are ecliptic-aligned 
	Returns:
		aei (n x 6 array): array with (x,y,z,vx,vy,vz) elements
	'''
	mu = SunGM if helio else SolarSystemGM


	M0 = np.array((epoch - keplerian[:,5])*(np.sqrt(mu)/np.power(keplerian[:,0],3./2)))
	#print(M0)

 
	E = solve_anomaly(np.array(keplerian[:,1]), M0)
	#return M0, E, keplerian[:,1]

	q_vec = q(E, keplerian[:,0], keplerian[:,1])

	Rot = R(keplerian[:,4], keplerian[:,3], keplerian[:,2])

	x = np.einsum('...ij,j...', Rot, q_vec)

	dqdt_vec = dqdt(E, keplerian[:,0], keplerian[:,1], mu)

	v = np.einsum('...ij,j...', Rot, dqdt_vec)

	xv = np.zeros_like(keplerian)

	xv[:,0:3] = x
	xv[:,3:] = v

	if not ecliptic:
		xv = rotate_to_ecliptic(xv)
	return xv



def dist_to_point(elements, epoch, element_type, point, helio = False, ecliptic = False):
	'''
	Computes the distance between all 6d elements and the 3d reference point. Computes coordinate changes as needed

	Args:
		elements (n x 6 array): Array with cartesian coordinates or Keplerian orbital elements
		epoch (float): epoch at which elements are defined, not really needed for Cartesian elements
		element_type (str): Either 'keplerian' or 'cartesian', defines which orbital element representation to use
		point (3d array): reference point
		helio (bool, default = False): Switches between barycentric and heliocentric elements (False is barycentric)
		ecliptic (bool, default = False): Use true if input elements are equatorially-aligned (ICRS)
	Returns:
		r (n x 1 array): array with distances

	'''
	if element_type == 'keplerian':
		xv = keplerian_to_cartesian(elements, epoch, helio, ecliptic)
	elif element_type == 'cartesian':
		xv = elements
	else:
		raise ValueError("Element type must either be keplerian or cartesian!")

	r = np.sqrt((xv[:,0] - point[0])**2 + (xv[:,1] - point[1])**2 + (xv[:,2] - point[2])**2)

	return r

def bary_to_helio(elements, element_type, epoch, sun_coordinates, ecliptic = False):
	'''
	Converts a set of barycentric elements (either Keplerian or cartesian) to heliocentric, returning them in the same fashion
	sun_coordinates must be a 6D vector in (AU,AU/yr)

	Args:
		elements (n x 6 array): Array with barycentric cartesian coordinates or Keplerian orbital elements 
		epoch (float): epoch at which elements are defined, not really needed for Cartesian elements
		element_type (str): Either 'keplerian' or 'cartesian', defines which orbital element representation to use
		sun_coordinates (3d array): coordinates of the sun at given epoch
		ecliptic (bool, default = False): Use true if input elements are equatorially-aligned (ICRS)
	Returns:
		n x 6 array in the same orbital element representation used as input, in the heliocentric frame
	'''
	if element_type == 'keplerian':
		xv = keplerian_to_cartesian(elements, epoch, False, ecliptic)
	elif element_type == 'cartesian':
		xv = elements
	else:
		raise ValueError("Element type must either be keplerian or cartesian!")

	
	new_elements = xv[:,] - sun_coordinates

	if element_type == 'keplerian':
		new_elements = cartesian_to_keplerian(new_elements, epoch, True, ecliptic)

	return new_elements

def helio_to_bary(elements, element_type, epoch, bary_coordinates, ecliptic = False):
	'''
	Converts a set of heliocentric elements (either Keplerian or cartesian) to barycentric, returning them in the same fashion
	bary_coordinates must be a 6D vector in (AU,AU/yr)

	Args:
		elements (n x 6 array): Array with heliocentric cartesian coordinates or Keplerian orbital elements 
		epoch (float): epoch at which elements are defined, not really needed for Cartesian elements
		element_type (str): Either 'keplerian' or 'cartesian', defines which orbital element representation to use
		bary_coordinates (3d array): coordinates of the barycenter at given epoch
		ecliptic (bool, default = False): Use true if input elements are equatorially-aligned (ICRS)
	Returns:
		n x 6 array in the same orbital element representation used as input, in the barycentric frame
	'''
	
	if element_type == 'keplerian':
		xv = keplerian_to_cartesian(elements, epoch, True, ecliptic)
	elif element_type == 'cartesian':
		xv = elements
	else:
		raise ValueError("Element type must either be keplerian or cartesian!")

	
	new_elements = xv[:,] - bary_coordinates

	if element_type == 'keplerian':
		new_elements = cartesian_to_keplerian(new_elements, epoch, False, ecliptic)

	return new_elements

def equatorial_to_galactic(vector):
	'''
	Transforms a 3d vector from equatorial coordinates to Galactic coordinates, following the transformations outlined in van Altena (2013)

	'''
	coslOm =   np.cos(np.pi *  32.93192/180)
	sinlOm =   np.sin(np.pi *  32.93192/180)
	cosalphG = np.cos(np.pi *  192.85948/180)
	sinalphG = np.sin(np.pi *  192.85948/180)
	cosdeltG = np.cos(np.pi *  27.12825/180)
	sindeltG = np.sin(np.pi *  27.12825/180)

	R_eq_to_gl = np.array([[-coslOm * sinalphG + cosalphG * sindeltG * sinlOm, cosalphG*coslOm+sinalphG*sindeltG*sinlOm,- cosdeltG * sinlOm], 
				[- cosalphG * coslOm * sindeltG - sinalphG * sinlOm, - coslOm * sinalphG * sindeltG + cosalphG*sinlOm,cosdeltG*coslOm], 
				[cosalphG*cosdeltG, cosdeltG*sinalphG,sindeltG]])

	return np.einsum('ij,...j',R_eq_to_gl,vector)

def galactic_to_equatorial(vector):
	'''
	Transforms a 3d vector from Galactic coordinates to equatorial coordinates, following the transformations outlined in van Altena (2013)
	'''
	coslOm =   np.cos(np.pi *  32.93192/180)
	sinlOm =   np.sin(np.pi *  32.93192/180)
	cosalphG = np.cos(np.pi *  192.85948/180)
	sinalphG = np.sin(np.pi *  192.85948/180)
	cosdeltG = np.cos(np.pi *  27.12825/180)
	sindeltG = np.sin(np.pi *  27.12825/180)

	R_gl_to_eq = np.linalg.inv(np.array([[-coslOm * sinalphG + cosalphG * sindeltG * sinlOm, cosalphG*coslOm+sinalphG*sindeltG*sinlOm,- cosdeltG * sinlOm], 
				[- cosalphG * coslOm * sindeltG - sinalphG * sinlOm, - coslOm * sinalphG * sindeltG + cosalphG*sinlOm,cosdeltG*coslOm], 
				[cosalphG*cosdeltG, cosdeltG*sinalphG,sindeltG]]))

	return np.einsum('ij,...j',R_gl_to_eq,vector)



def invariable_to_ecliptic(vector):
	'''
	Transforms a 3d vector from invariable plane coordinates to ecliptic coordinates, following Souami & Souchay (2012)
	'''

	cosphi = np.cos(1.587 * np.pi/180)
	sinphi = np.sin(1.587 * np.pi/180)
	cosLan = np.cos(107.6 * np.pi/180)
	sinLan = np.sin(107.6 * np.pi/180)

	R_inv_to_ecl = np.array([[cosLan, -cosphi * sinLan, sinphi], [sinLan,cosphi*cosLan,-cosLan*sinphi],[0,sinphi,cosphi]])


	return np.einsum('ij,...j', R_inv_to_ecl, vector)



from .decam import *
import os 
import astropy.table as tb
from astropy.wcs import WCS

try:
	import pixmappy
	pix = True
except ModuleNotFoundError:
	pix = False

class DESExposure(DECamExposure):
	'''
	DECamExposure with additional quantities from DES analysis. Examples: astrometry covariances and exposure completeness
	'''
	def __init__(self, expnum, ra, dec, mjd_mid, band, m50 = None, c = None, k = None, cov = None):
		'''
		Initialization class

		Arguments:
		- expnum: Exposure number 
		- ra: R.A. pointing (degrees)
		- dec: Declination pointing (degrees)
		- mjd_mid: midpoint time of the exposure (mjd)
		- band: exposure filter
		- m50: magnitude of 50% completeness
		- c: efficiency for completeness
		- k: transition slope for completeness
		- cov: astrometric covariance matrix
		'''
		DECamExposure.__init__(self,expnum, ra, dec, mjd_mid, band)
		self.m50 = m50 
		self.c = c
		self.k = k
		#cov is in mas^2
		xx = cov[0]
		yy = cov[1]
		xy = cov[2]
		self.cov = np.array([[xx, xy], [xy, yy]]) * (1/3600000**2)
		

	def probDetection(self, m):
		'''
		Computes the detection probability of something with magnitude m in this exposure
		
		Arguments:
		- m: magnitude (numpy array or float)
		
		Returns:
		- detection probability, same shape as m
		'''
		if self.m50 != None:
			return self.c/(1. + np.exp(self.k * (m - self.m50)))
		else:
			return np.zeros_like(m)

	def getPixelCoordinates(self, ra, dec, ccd, c = 0.61):
		'''
		Given a list of RAs and Decs on a given CCD, returns the x,y coordinates of the detections
		using the pixmappy WCSs

		Arguments:
		- ra: list of R.A.s (degrees)
		- dec: list of Decs (degrees)
		- ccd: list of CCD numbers for WCS mapping
		'''
		try:
			return self.wcs[ccd].toPix(ra, dec, c = c)
		except:
			raise ValueError("No CCD solution for {}/{}!".format(self.expnum, ccd))

	def createWCSDict(self, pmc = None):
		'''
		Uses pixmappy to grab the CCD WCS solution

		Arguments:
		- pmc: pixmappy DESMaps() instance (optional)
		'''
		if not pix:
			raise ModuleNotFoundError("You need pixmappy for this function!")
		self.wcs = {}
		self.fullwcs = True
		if pmc == None:
			from pixmappy import DESMaps
			pmc = DESMaps()

		for i in range(1,63):
			if i != 61:
				try:
					self.wcs[i] = pmc.getDESWCS(self.expnum, i)
				except:
					self.wcs[i] = None
					if i != 31 or i != 2:
						self.fullwcs = False
	
	def getWCS(self, ccdnum, pmc = None):
		'''
		Uses pixmappy to grab the CCD WCS solution

		Arguments:
		- ccdnum: CCD number for solution
		- pmc: pixmappy DESMaps() instance (optional)

		'''
		if not pix:
			raise ModuleNotFoundError("You need pixmappy for this function!")

		if pmc == None:
			from pixmappy import DESMaps
			pmc = DESMaps()

		return pmc.getDESWCS(self.expnum, ccdnum)

	def makeWCS(self, expinfo):
		'''
		Generates a WCS dictionary using astropy.wcs and the wcs table for the exposure. 
		Note that these are not the pixmappy solutions, so are less accurate

		Arguments:
		- expinfo: table of WCS information for the exposures coming from FITS headers
		'''
		self.wcs_db = {}
		for i in expinfo:
			d = {k:i[k] for k in i.keys()}
			self.wcs_db[i['CCDNUM']] = WCS(header=d)

	def samplePosError(self, shotnoise, n_samples):
		'''
		Finds n_samples position errors given the shot noise for each value and the
		atmospheric turbulence error matrix

		Arguments:
		- shotnoise: shot noise error in degrees
		- n_samples: number of samples from the covariance matrix to be drawn
		'''

		#shot_err = shotnoisedist.sample(n_samples)

		atm_err = np.random.multivariate_normal(np.array([0,0]), self.cov, n_samples)

		gauss = np.random.normal(size=(n_samples, 2))

		for i in range(n_samples):
			gauss[i] *= shotnoise[i]

		err = gauss + atm_err

		err[:,0] *= np.cos(self.dec * np.pi/180)

		return err

	

class DES(Survey):
	'''
	Full DES survey based on one of the release files. Examples: y4a1, y4a1c, y6a1, y6a1c
	Main change between this and Survey is the support for completeness and atmospheric turbulence terms in the exposures
	'''
	def __init__(self, release):
		'''
		Initialization function

		Arguments:
		- release: name of the release, all data files need be of the form release.EXTENSION
		'''
		orbdata = os.getenv('DESDATA')

		self.release = release
		track ='{}/{}.exposures.positions.fits'.format(orbdata, self.release)
		corners = '{}/{}.ccdcorners.fits.gz'.format(orbdata, self.release)
		
		exp = tb.Table.read(track, 1)
		try:
			Survey.__init__(self, exp['EXPNUM'], exp['RA'], exp['DEC'], exp['MJD_MID'], exp['BAND'], track = track, corners = corners)
		except:
			Survey.__init__(self, exp['expnum'], exp['ra'], exp['dec'], exp['mjd_mid'], exp['filter'], track = track, corners = corners)
		self.exp = exp

		if 'm50' in exp.keys():
			self.m50 = exp['m50']
			self.c = exp['c']
			self.k = exp['k']
		else:
			self.m50 = len(exp) * [None]
			self.c = len(exp) * [None]
			self.k = len(exp) * [None]

		if 'cov' in exp.keys():
			self.cov = exp['cov']
		else:
			self.cov = len(exp) * [np.array([[0,0],[0,0]])]


	def createExposures(self):
		'''
		Makes a dictionary of DESExposures, which can be accessed by calling DES[exposure]
		'''
		self.exposures = {}
		for ra, dec, mjd, n, b, m50, c, k, cov in zip(self.ra, self.dec, self.mjd, self.expnum, self.band, self.m50, self.c, self.k, self.cov):
			self.exposures[n] = DESExposure(n, ra, dec, mjd, b, m50, c, k, cov)

	def observePopulation(self, population, lightcurve = None, keepall = False):
		'''
		Uses the population's magnitudes to check if a detection is observable given each exposure completeness

		Arguments:
		- population: Population object to be observed, requires magnitudes of each object
		- lightcurve: list (or single) of lightcurve objects to these can be applied for the objects. If a list is provided, make sure len(lightcurve) == len(population)
		- keepall: boolean, will split orbits between detections and non-detections if True
		'''
		try:
			population.observations
		except:
			raise AttributeError("Population has no observations")

		try:
			self.exposures
		except:
			raise AttributeError("Survey does not have a list of DESExposures")

		exp = tb.Table()
		exp['EXPNUM'] = self.expnum
		exp['BAND'] = self.band
		exp['m_50'] = self.m50
		exp['C'] = self.c 
		exp['K'] = self.k 
		exp['MJD'] = self.mjd

		#we don't need all exposures - should save memory
		exp = exp[np.isin(exp['EXPNUM'], population.observations['EXPNUM'])]

		obs = tb.join(population.observations, exp)

		# remove what doesn't have observations before joining!
		mags = population.mag_obs[np.isin(population.mag_obs['ORBITID'], np.unique(obs['ORBITID']))]
		
		obs = tb.join(obs, mags)

		population.obs = obs

		if lightcurve is not None:
			population.generateLightCurve(lightcurve)

		del exp, mags


		population.obs['DETPROB'] = population.obs['C']/(1. + np.exp(population.obs['K'] * (population.obs['MAG'] - population.obs['m_50'])))


		population.obs['RANDOM'] = np.random.rand(len(population.obs))

		del population.obs['m_50', 'C', 'K']

		if not keepall:
			ndet = population.obs[population.obs['DETPROB'] < population.obs['RANDOM']]
			obs = population.obs[population.obs['DETPROB'] > population.obs['RANDOM']]
			population.ndet = ndet
		else:
			obs = population.obs

		population.detections = obs









import numpy as np 
import numba

from numba.typed import List

# Uncomment the following line to print out the compilation steps
#cc.verbose = True

#@cc.export('compute_arccut', 'f8(f8[:])')

@numba.jit('f8(f8[:])', nopython=True,)
def compute_arccut(times):
	'''
	Computes ARCCUT, the time between the first and last detection dropping one night of detection
	Arguments:
	- times: list of times, must be in DAYS
	'''
	arccut = 0.
	t1 = np.min(times)
	t2 = np.max(times)
	t1a = t2
	t2a = t1

	n = len(times)

	for i in range(n):
	  if times[i] - t1 > 0.7 and times[i] < t1a:
	  	t1a = times[i]
	  if t2 - times[i] > 0.7 and times[i] > t2a:
	  	t2a = times[i]

	arccut = min(t2 - t1a, t2a - t1)

	return arccut

#@cc.export('compute_nunique', 'i8(f8[:])')

@numba.jit('i8(f8[:])',nopython=True,)
def compute_nunique(times):
	'''
	Computes NUNIQUE, the number of unique nights in which we have observations from an object

	Arguments:
	- times: list of times, must be in DAYS
	'''
	nunique = 0

	n = len(times)

	unique = True 

	for i in range(n):
		unique = True
		for j in range(0, i):
			if abs(times[j] - times[i]) < 0.4:
				unique = False 
		if unique:
			nunique += 1

	return nunique

#@cc.export('compute_triplet', 'b1(f8[:], f8)')

@numba.jit('b1(f8[:], f8)',nopython=True,)
def compute_triplet(times, thresh):
	'''
	Computes the time difference between triplets, returns if a triplet is formed

	Arguments:
	- times: list of times, must be in DAYS
	- thresh: threshold time for pairs
	'''
	first_pair =  999.
	second_pair = 999.
	det = List()

	det.append(times[0])

	n = len(times)

	for i in range(n-1):
		if times[i + 1] - times[i] > 0.4:
			det.append(times[i+1])
	
	n = len(det)
	for i in range(1,n-1):
		if det[i + 1] - det[i] < first_pair and det[i] - det[i-1] < second_pair:
			first_pair = det[i + 1] - det[i]
			second_pair = det[i] - det[i-1]

	if first_pair < thresh and second_pair < thresh:
		return True
	else:
		return False

#@cc.export('find_triplet_time', 'f8[:](f8[:])')

@numba.jit('f8[:](f8[:])',nopython=True,)
def find_triplet_time(times):
	'''
	Computes the minimum triplet time for both pairs of an object

	Arguments:
	- times: list of times, must be in DAYS

	'''
	first_pair =  999.
	second_pair = 999.
	det = List()

	det.append(times[0])

	n = len(times)

	for i in range(n-1):
		if times[i + 1] - times[i] > 0.4:
			det.append(times[i+1])
	
	n = len(det)
	for i in range(1,n-1):
		if det[i + 1] - det[i] < first_pair and det[i] - det[i-1] < second_pair:
			first_pair = det[i + 1] - det[i]
			second_pair = det[i] - det[i-1]

	return np.array([first_pair, second_pair])




'''
Series of convenience functions for working with magnitudes
'''
import numpy as np 
import astropy.table as tb 

from scipy.optimize import minimize

alpha_i  =  1.492983987794142 
beta_i   = -0.12310383521046273 
alpha_z  =  1.65452092361025 
beta_z   = -0.13315208762781655



def generate_colors(gr):
	'''
	Generates g - i and g - z colors from a nominal g - r colors, according to a linear fit 

	Arguments:
	- gr: array of gr colors
	'''
	gi = alpha_i*gr + beta_i
	gz = alpha_z*gr + beta_z
	return gi , gz  

def shot_noise(mag, band):
	'''
	Return mean shot noise error expected as a function of magnitude in a given band

	Arguments:
	- mag: magnitudes for the shot noise error (array)
	- band: single band, must be one of 'g', 'r', 'i', 'z'
	'''
	shot_noise_amp = {'g' : -10.237155096761413, 'r' : -10.306198159582513, 'i' : -10.182735894436972, 'z' : -10.015274570909892}

	return np.power(10, shot_noise_amp[band] + 0.4*mag)/3600

def detprob_logit(m, params):
    '''
    logit function
    params = (m50, k, c)

    Arguments:
    - m: magnitude argument
    - params: tuple with m50, k and c
    '''
    m50, k, c = params
    logit = c/(1+np.exp(k*(m-m50)))
    return logit

def detprob_double(m, params):
    '''
    double logit function
    params = (m0, k1, k2, c)

     Arguments:
    - m: magnitude argument
    - params: tuple with m0, k1, k2 and c
    '''
    m50, k1, k2, c = params
    logit = c/(1+np.exp(k1*(m-m50)))/(1+np.exp(k2*(m-m50)))
    return logit

def detprob_piecewise(x, params, split):
	'''
	Piecewise logit function
	params = (r50_1, r50_2, kappa1, kappa2, c)

	Arguments:
	- m: magnitude argument
	- params: tuple with r50_1, c, r50_2, kappa1 and kappa2 
	'''

	r50_1, r50_2, kappa1, kappa2, c = params
	p1 = lambda x : detprob_logit(x, [r50_1, kappa1, c])
	p2 = lambda x : detprob_logit(x, [r50_2, kappa2, c])
	return np.piecewise(x,  [x<split, x>=split], [p1, p2])


def minusLogP(params, mdet, mnon, res_collect, detprob):
    '''
    Takes logit parameters and list of detected and non-detected magnitudes. 
    Returns negative log of pdf


    Arguments:
    - params: tuple of parameters for the fit
    - mdet: array of detected magnitudes
    - mnon: array of missed magnitudes
    - res_collect: list of previous failed results
    - detprob: functional form (eg detprob_logit or detprob_double)
    '''

    if params[-1] > 1.:
        # this ensures that c cannot continue to rise above 1 by referencing 
        # the value of the previous trial in the optimizer
        res_collect.append(res_collect[-1] + (params[-1] - 1.)*1e5)
        return res_collect[-1] + 1e7
    elif params[-1] <= 0.:
        return res_collect[-1] + 1e7
    else:
        pdet = detprob(mdet,params)
        pnon = detprob(mnon,params)    
        result = np.sum(np.log(pdet))
        result += np.sum(np.log(1-pnon))
        return -result

def find_m50(m_det, m_miss, detprob = detprob_logit, init = (23,0.9,5)):
	'''
	Fits the provided function by taking the product of the detected probabilities and 1-non-detection probabilities

	Arguments:
	- m_det: array of detected magnitudes
    - m_non: array of missed magnitudes
    - detprob: functional form (eg detprob_logit or detprob_double)
    - init: tuple of initialization values, must match the arguments of detprob
	'''
	results_collector = [0]
	return minimize(minusLogP, init, method='Powell', args=(m_det, m_miss, results_collector, detprob), tol=1e-3)


def magnitude_physical(albedo, diameter, band):
	'''
	Finds the absolute magnitude of an object given an albedo in a certain band and a diameter (in km)
	Assumes phi(alpha) = 1
	Solar AB magnitudes from Willmer 2018 ApJS 236 47 (https://iopscience.iop.org/article/10.3847/1538-4365/aabfdf)

	Arguments:
	- albedo: albedo value
	- diameter: diameter of the object in km
	- band: one of 'g', 'r', 'i', 'z', 'Y' for absolute magnitude look-up
	'''

	solar_m = {'g' : -26.52, 'r' : -26.96, 'i' : -27.05, 'z' : -27.07, 'Y' : -27.07}

	log_term = albedo * diameter * diameter/9e16  

	return solar_m[band] - 2.5*np.log10(log_term)


def size_from_mag(albedo, magnitude, band):
	'''
	Finds the diameter (in km) of an object given an albedo in a certain band and an absolute magnitude 
	Assumes phi(alpha) = 10
	Solar AB magnitudes from Willmer 2018 ApJS 236 47 (https://iopscience.iop.org/article/10.3847/1538-4365/aabfdf)

	Arguments:
	- albedo: albedo value
	- magnitude: absolute magnitude of the object
	- band: one of 'g', 'r', 'i', 'z', 'Y' for absolute magnitude look-up

	'''

	solar_m = {'g' : -26.52, 'r' : -26.96, 'i' : -27.05, 'z' : -27.07, 'Y' : -27.07}

	power = np.power(10, 0.4*(solar_m[band] - magnitude))  

	return np.sqrt(9e16 * power/albedo)


class BaseLightCurve:
	'''
	Base method for lightcurves, returning no shift independently of the phase
	'''
	def __init__(self):
		'''
		Initialization function. Doesn't do anything
		'''
		pass
	def _sample(self, times):
		'''
		Samples this lightcurve

		Argument:
		- times: array of times
		'''
		return np.zeros_like(times)
	def __call__(self, x):
		return self._sample(x)

class RandomPhaseLightCurve(BaseLightCurve):
	'''
	Returns a randomly sampled sinusoidal lightcurve with a given peak-to-peak amplitude
	'''
	def __init__(self, amplitude):
		'''
		Initialization function

		Argument:
		- amplitude: amplitude of the light curve (will oscillate between +- amplitude/2)
		'''
		self.amplitude = amplitude

	def _sample(self, times):
		'''
		Samples the light curve

		Argument:
		- times: array of times
		'''
		phases = np.random.sample(times.shape)*2 *np.pi 
		return self.amplitude * np.sin(phases)/2

class SinusoidalLightCurve(BaseLightCurve):
	'''
	Given a certain known lightcurve with a given amplitude, period and phase, returns each point of the lightcurve
	'''
	def __init__(self, amplitude, period, phase):
		'''
		Initialization function

		Argument:
		- amplitude: amplitude of the light curve (will oscillate between +- amplitude/2)
		- period: period of the oscillation
		- phase: phase of the oscillation, in radians
		'''
		self.amplitude = amplitude
		self.period = period
		self.phase = phase

	def _sample(self, times):
		'''
		Samples the light curve

		Argument:
		- times: array of times
		'''
		return self.amplitude * np.sin(2 * np.pi * times/self.period + self.phase)






### Following pixmappy's setup 
import glob, os  


try:
    from setuptools import setup
    import setuptools
    print("Using setuptools version",setuptools.__version__)
except ImportError:
    from distutils.core import setup
    import distutils
    print("Using distutils version",distutils.__version__)


dependencies = ['numpy', 'astropy', 'scipy', 'numba', 'spacerocks', 'rich']

with open('README.md') as file:
	long_description = file.read()


## ignoring version stuff for now 

data = glob.glob(os.path.join('data', '*'))

dist = setup(
	name = "DESTNOSIM",
	version = "1.3.1",
	author = "Pedro Bernardinelli",
	author_email = "pedrobe@sas.upenn.edu",
	description = "Python module for simulating DES TNO observations",
	long_description = long_description,
	long_description_content_type='text/markdown',
	license = "BSD License",
	url = "https://github.com/bernardinelli/DESTNOSIM",
	packages = ['destnosim', 'destnosim.des', 'destnosim.tno'],
	package_data = {'destnosim' : data},
	install_requires = dependencies,)



'''
Sampler classes for many different distributions. These are all based on the `BaseDistribution` class, 
and by subclassing it any distribution can be implemented. The JointDistribution allows one to construct a distribution of 
the form 
	alpha p_1(x) + (1-alpha) p_2(x),
where 0 <= alpha <= 1 and p_1, p_2 are two distributions

'''
import numpy as np
from .inv_sample import *

class BaseDistribution:
	'''
	Base class that all others inherit from. It corresponds to a delta function at zero. 
	'''
	def __init__(self):
		'''
		Initialization function. Doesn't do anything
		'''
		pass
	def __add__(self, other):
		return JointDistribution(self, other, 0.5)
	def __len__(self):
		return 0

	def sample(self, n):
		'''
		Samples from this distribution

		Arguments:
		- n: number of samples
		'''
		return np.zeros(n)


class AnalyticDistribution(BaseDistribution):
	'''	
	Given an analytic probability density function, samples from it using the inverse transform sampling algorithm from inv_sample.py
	'''
	def __init__(self, x_min, x_max, function, n_samp = 1000, n_bins = 100):
		'''
		Initialization function

		Arguments:
		- x_min: minimum value for the distribution
		- x_max: maximum value for the distribution
		- function: functional form of the distribution (eg a lambda function)
		- n_samp: number of samples for the inv_sample code
		- n_bins: number of bins for for the inv_sample code
		'''
		self.f = function
		self.x_min = x_min
		self.x_max = x_max
		self.n_samp = n_samp
		self.n_bins = n_bins
		self._constructSampler() 

	def _constructSampler(self):
		x = np.linspace(self.x_min, self.x_max, self.n_samp)
		y = self.f(x)
		self.invCDF = inverse_cdf(x, y, self.n_bins)

	def sample(self, n):
		'''
		Samples from this distribution

		Arguments:
		- n: number of samples
		'''
		r = np.random.rand(n)
		return self.invCDF(r)

	def __add__(self, other):
		x_min = np.max(self.x_min, other.x_min)
		x_max = np.min(self.x_max, other.x_max)
		f = lambda x : self.f(x) + other.f(x)
		n_samp = np.max(self.n_samp, other.n_samp)
		n_bins = np.max(self.n_bins, other.n_bins)

		return AnalyticDistribution(x_min, x_max, f, n_samp, n_bins)

class DistributionFromHistogram(AnalyticDistribution):
	'''
	Samples from the provided histogram using an inverse sampling algorithm
	'''
	def __init__(self, bins, pdf):
		'''
		Initialization function

		Arguments:
		- bins: bin edges of the histogram (numpy style)
		- pdf: value of the histogram at the bins (numpy style)
		'''
		self.bins = bins
		self.pdf = pdf 

		self._constructSampler()

	def _constructSampler(self):
		self.invCDF = inverse_cdf_histogram(bin_edges = self.bins, hist = self.pdf)

		
class DeltaFunction(BaseDistribution):
	'''
	Delta function centered at the provided location
	'''
	def __init__(self, loc):
		'''
		Initialization function

		Arguments:
		- loc: location of the delta function peak
		'''
		self.loc = loc
		pass
	def sample(self, n):
		'''
		Samples from this distribution

		Arguments:
		- n: number of samples
		'''

		return self.loc * np.ones(n)

class PowerLaw(BaseDistribution):
	'''
	Power law of the form x^slope, defined between x_min and x_max
	'''
	def __init__(self, slope, x_min, x_max):
		'''
		Initialization function

		Arguments:
		- slope: slope - 1 of the power law (for numpy reasons)
		- x_min: minimum value for the distribution
		- x_max: maximum value for the distribution
  
		'''
		self.slope = slope
		self.x_min = x_min
		self.x_max = x_max
		self.scale = x_max - x_min
		self.norm = self._normalization()

	def sample(self, n):
		'''
		Samples from this distribution

		Arguments:
		- n: number of samples
		'''

		samp = np.random.power(self.slope + 1, size=n)
		return self.x_min + self.scale * samp
	
	def _normalization(self):
		return (self.x_max**(self.slope + 1) - self.x_min**(self.slope + 1))/(self.slope + 1)

class JointDistribution(BaseDistribution):
	'''
	This returns a distribution that approximately samples alpha dist_1 + (1-alpha) dist_2
	This is not exact, as it forces the proportionalities for the sampling
	'''
	def __init__(self, distribution_1, distribution_2, alpha):
		'''
		Initialization function

		Arguments:
		- distribution_1: first distribution object, 100*alpha% of the samples will come from it 
		- distribution_2: second distribution object, 100*(1-alpha)% of the samples will come from it 
		- alpha: relative size between distributions
		'''
		self.dist_1 = distribution_1
		self.dist_2 = distribution_2
		if alpha > 1 or alpha < 0:
			raise ValueError("The value of alpha must be between 0 and 1")
		self.alpha = alpha

	def sample(self, n):
		'''
		Samples from this distribution

		Arguments:
		- n: number of samples
		'''

		n_frac = int(self.alpha*n)
		samp1 = self.dist_1.sample(n_frac)
		samp2 = self.dist_2.sample(n - n_frac)

		samp = np.append(samp1, samp2)
		perm = np.random.permutation(n)
		return samp[perm]

class Uniform(PowerLaw):
	'''
	Creates a uniform distribution between two different values
	'''
	def __init__(self, x_min, x_max):
		'''
		Initialization function
		
		Arguments:
		- x_min: minimum value for the distribution
		- x_max: maximum value for the distribution
		'''
		PowerLaw.__init__(self, 0, x_min, x_max)

class GaussianDistribution(BaseDistribution):
	def __init__(self, mu, sigma):
		self.mu = mu 
		self.sigma = sigma
	def sample(self, n):
		return np.random.normal(size=n, loc=self.mu, scale=self.sigma)
		

class Logarithmic(Uniform):
	'''
	Function for logarithmically distributed samples between two values
	'''
	def __init__(self, x_min, x_max):
		'''
		Initialization function
		
		Arguments:
		- x_min: minimum value for the distribution
		- x_max: maximum value for the distribution
		'''
		self._sampler = Uniform(np.log(x_min), np.log(x_max))
		self.x_min = x_min
		self.x_max = x_max

	def sample(self, n):
		'''
		Samples from this distribution

		Arguments:
		- n: number of samples
		'''
		return np.exp(self._sampler.sample(n))

class BrokenPowerLaw(AnalyticDistribution):
	'''
	Creates a broken power law with two different slopes, transitioning at a certain break point
	'''
	def __init__(self, slope_1, slope_2, x_min, x_max, x_break, x_norm = None):
		'''
		Initialization function

		Arguments:
		- slope_1: first slope, for x < x_break
		- slope_2: second slope, for x > x_break
		- x_min: minimum value for the distribution
		- x_max: maximum value for the distribution
		- x_break: break point of the distribution
		- x_norm: normalization for the x values, for numerical stability
		'''
		self.x_break = x_break
		self.slope_1 = slope_1
		self.slope_2 = slope_2
		if x_norm == None:
			x_norm = (x_min + x_max)/2
		self.x_norm = x_norm

		self.f = lambda x : np.piecewise(x, [ x <= self.x_break, x > self.x_break], [lambda x: np.power(x - x_norm, self.slope_1), 
														lambda x: np.power(x - x_norm, self.slope_2) * np.power(self.x_break - x_norm, self.slope_1 - self.slope_2)])
	
		AnalyticDistribution.__init__(self, x_min, x_max, self.f)

class DoublePowerLaw(AnalyticDistribution):
	'''
	Creates a double power law with two different slopes that become equal at a certain point x_eq
	'''

	def __init__(self, slope_1, slope_2, x_min, x_max, x_eq, x_shift):
		'''
		Initialization function

		Arguments:
		- slope_1: first slope, for x < x_break
		- slope_2: second slope, for x > x_break
		- x_min: minimum value for the distribution
		- x_max: maximum value for the distribution
		- x_eq: point where the two power laws are equal
		- x_shift: the power laws use (x - x_shift), this value defines the shift point
		'''

		self.x_eq = x_eq
		self.slope_1 = slope_1
		self.slope_2 = slope_2
		self.x_shift = x_shift
		self.x_min = x_min
		self.x_max = x_max
		c = np.power(10, (slope_2 - slope_1)*(x_eq - x_shift))
		self.c = c
		self.f = lambda x : (1 + c)/(np.power(10, -slope_1 * (x - x_shift)) + c * np.power(10, -slope_2 * (x - x_shift)))


		AnalyticDistribution.__init__(self, x_min, x_max, self.f)

class RollingPowerLaw(AnalyticDistribution):
	'''
	Creates a rolling power law with two different slopes, one linear and one quadratic in the exponential
	'''
	def __init__(self, slope_1, slope_2, x_min, x_max, x_shift):
		'''
		Initialization function

		Arguments:
		- slope_1: first slope, for x < x_break
		- slope_2: second slope, for x > x_break
		- x_min: minimum value for the distribution
		- x_max: maximum value for the distribution
		- x_shift: the power laws use (x - x_shift), this value defines the shift point
		'''
		self.slope = slope
		self.derivative = derivative
		self.x_shift = x_shift
		self.x_min = x_min
		self.x_max = x_max
		self.f = lambda x : np.power(10, slope * (x - x_shift) + derivative * (x - x_shift)**2 )
		AnalyticDistribution.__init__(self, x_min, x_max, self.f)


class BrownDistribution(AnalyticDistribution):
	'''
	Brown distribution (Brown 2001) of the form sin(x) * exp(-(x-x_center)**2/2 sigma**2)
	Assumes x is in degrees!
	'''
	def __init__(self, x_min, x_max, sigma, x_center):
		'''
		Initialization function

		Arguments:
		- x_min: minimum value for the distribution
		- x_max: maximum value for the distribution
		- sigma: sigma of the Gaussian term
		- x_center: center of the Gaussian term
		'''
		self.sigma = sigma
		self.f = lambda x : np.sin(x * np.pi/180) * np.exp(- ((x - x_center)/sigma)**2/2.)
		AnalyticDistribution.__init__(self, x_min, x_max, self.f)

class SinusoidalDistribution(AnalyticDistribution):
	'''
	Sinusoidal distribution
	'''
	def __init__(self, x_min, x_max):
		'''
		Initialization function
		
		Arguments:
		- x_min: minimum value for the distribution
		- x_max: maximum value for the distribution
		'''
		self.f = lambda x : np.sin(x * np.pi/180)
		AnalyticDistribution.__init__(self, x_min, x_max, self.f)

class RayleighDistribution(AnalyticDistribution):
	'''
	Rayleigh distribution of the form (x/sigma^2) * exp(- (x-x_center)^2 / 2 sigma^2)
	'''
	def __init__(self, x_min, x_max, sigma, x_center):
		'''
		Initialization function

		Arguments:
		- x_min: minimum value for the distribution
		- x_max: maximum value for the distribution
		- sigma: sigma of the Gaussian term
		- x_center: center of the Gaussian term
		'''

		self.sigma = sigma
		self.f = lambda x : (x /(sigma)**2)* np.exp(- ((x - x_center)/sigma)**2/2.)
		AnalyticDistribution.__init__(self, x_min, x_max, self.f)

class FunctionalUniform(Uniform):
	'''
	Applies a given function to a sample of the Uniform distribution
	'''
	def __init__(self, x_min, x_max, function):
		'''
		Initialization function

		Arguments:
		- x_min: minimum value for the distribution
		- x_max: maximum value for the distribution
		- function: function to be applied to the uniform samples
		'''

		self.f = function 
		self.uniform = Uniform(x_min, x_max)
		Uniform.__init__(self, x_min, x_max)

	def sample(self, n):
		'''
		Samples from this distribution

		Arguments:
		- n: number of samples
		'''

		s = self.uniform.sample(n)
		return self.f(s)



from .decam import *
from ..tno.magnitude import * 

class PointingGroup(DECamExposure):
	def __init__(self, ra, dec, exposure_list, expnum1, expnum2, mjd1, mjd2, band, m25 = None, k1 = None, k2 = None, c = None):
		self.exposure_list = exposure_list
		self.expnum1 = expnum1
		self.expnum2 = expnum2
		self.ra = ra
		if self.ra > 180:
			self.ra -= 360
		self.dec = dec
		self.mjd1 = mjd1
		self.mjd2 = mjd2
		self.band = band
		self.corners = {}

		self.m25 = m25 
		self.k1 = k1 
		self.k2 = k2 
		self.c = c

	def collectExposures(self, survey):
		self.exposures = {} 

		for i in self.exposure_list:
			try:
				self.exposures[i] = survey[i]
			except:
				print(f'EXPNUM {i} not in survey')

	def probDetection(self, m):
		'''
		Computes the detection probability of something with magnitude m in this pointing group
		
		Arguments:
		- m: magnitude (numpy array or float)
		
		Returns:
		- detection probability, same shape as m
		'''
		if self.m25 != None:
			term1 = (1. + np.exp(self.k1 * (m - self.m25))) 
			term2 = (1. + np.exp(self.k2 * (m - self.m25)))
			return self.c/(term1 * term2)
		else:
			return np.zeros_like(m)

	def calculateRate(self, population, expmin = 50, ccdmask = [2,61], compute_mag = False):
		'''
		Uses the positions of the population to determine the rates, and also checks whether the source would be
		inside a functional CCD for the duration of the stare

		Arguments:
		- population: Population object from tno/population containing the input orbits and detections
		- expmin (int): minimum number of exposures the source has to be inside the CCD
		- ccdmask (list): list of masked CCDs
		'''
		try:
			population.obs
		except:
			raise AttributeError("Population has no obs attribute. Perhaps you need to call survey.observePopulation first?")

		## first, reduce population to this PG

		pop_pg = population.obs[np.isin(population.obs['EXPNUM'], self.exposure_list)]
		
		if len(pop_pg) == 0:
			return tb.Table()
		## see which objects are actually inside a searched CCD
		unique_ccds, counts = np.unique(pop_pg['ORBITID', 'CCDNUM'], return_counts = True)
		unique_ccds = tb.Table(unique_ccds)
		unique_ccds['COUNTS'] = counts 
		unique_ccds = unique_ccds[(unique_ccds['COUNTS'] > expmin) & (unique_ccds['CCDNUM'] != 0)]
		unique_ccds = unique_ccds[np.isin(unique_ccds['CCDNUM'], ccdmask, invert = True)]

		## go to rates 
		pop_inccd = pop_pg[np.isin(pop_pg['ORBITID'], unique_ccds['ORBITID'])]

		# this will be in degrees
		pop_inccd['THETA_X'], pop_inccd['THETA_Y'] = self.gnomonicProjection(pop_inccd['RA'], pop_inccd['DEC'])

		# don't see a way to avoid this for loop yet
		pop_inccd.add_index('ORBITID')

		rates = []
		if compute_mag:
			mags = []
		for i in np.unique(pop_inccd['ORBITID']):
			obj = pop_inccd.loc[i]
			deltax = obj['THETA_X'][np.argmax(obj['EXPNUM'])] - obj['THETA_X'][np.argmin(obj['EXPNUM'])]
			deltay = obj['THETA_Y'][np.argmax(obj['EXPNUM'])] - obj['THETA_Y'][np.argmin(obj['EXPNUM'])]

			r = np.sqrt(deltax**2 + deltay**2)/(self.exposures[np.max(obj['EXPNUM'])].mjd - self.exposures[np.min(obj['EXPNUM'])].mjd)

			rates.append(r)
			if compute_mag:
				mags.append(np.mean(obj['MAG']))

		t = tb.Table()
		t['ORBITID'] = np.unique(pop_inccd['ORBITID'])
		t['RATE'] = np.array(rates) * (3600 / 0.263)
		if compute_mag:
			t['MAG'] = mags
		#t = tb.join(t, pop_pg)
		return t 




class DEEP(Survey):
	def __init__(self, release):
		'''
		Initialization function

		Arguments:
		- release: name of the release, all data files need be of the form release.EXTENSION
		'''
		orbdata = os.getenv('DESDATA')

		self.release = release
		track ='{}/{}.exposures.positions.fits'.format(orbdata, self.release)
		corners = '{}/{}.ccdcorners.fits.gz'.format(orbdata, self.release)
		pointing_group = '{}/{}.pointing.fits'.format(orbdata, self.release)
		
		exp = tb.Table.read(track, 1)
		try:
			Survey.__init__(self, exp['EXPNUM'], exp['RA'], exp['DEC'], exp['MJD_MID'], exp['BAND'], track = track, corners = corners)
		except:
			Survey.__init__(self, exp['expnum'], exp['ra'], exp['dec'], exp['mjd_mid'], exp['filter'], track = track, corners = corners)
		self.exp = exp
		## ensure keys are capitalized
		for i in self.exp.keys():
			self.exp.rename_column(i, i.upper())
		if 'FILTER' in self.exp.keys():
			self.exp.rename_column('FILTER', 'BAND')

		self.pg = tb.Table.read(pointing_group)

		self.params = [self.pg.meta['R50_1'], self.pg.meta['R50_2'], self.pg.meta['KAPPA1'], self.pg.meta['KAPPA2'], self.pg.meta['C']] 
		self.split = self.pg.meta['SPLIT']


	def __getitem__(self, key):
		'''
		Allows DECamExposures to be accessed by indexing the Survey object
		'''
		try:
			return self.exposures[key]
		except AttributeError:
			try:
				return self.pointing_groups[key]
			except AttributeError:
				raise AttributeError("Survey does not have a list of DECamExposures or PointingGroups!")
		except KeyError:
			try:
				return self.pointing_groups[key]
			except:
				raise KeyError("Exposure or PointingGroup {} not in survey".format(key))


	def createPointingGroup(self):
		''' 
		Creates the dictionary of pointing groups for the survey
		'''

		try:
			self.exposures
		except:
			self.createExposures()

		self.pointing_groups = {} 
		for i in self.pg:
			exp_list = self.exp[(self.exp['EXPNUM'] >= i['EXPNUM1']) & (self.exp['EXPNUM'] < i['EXPNUM2'])]['EXPNUM']
			self.pointing_groups[i['POINTING_GROUP']] = PointingGroup(i['RA_PG'], i['DEC_PG'], exp_list, i['EXPNUM1'], i['EXPNUM2'],
																	  i['MJD1'], i['MJD2'], i['BAND'], i['m25'], i['k1'], i['k2'], i['c'])
		for i in self.pointing_groups:
			self.pointing_groups[i].collectExposures(self)

	def probRate(self, rate):
		'''
		Computes the probability that this object would be detected by kbmod in this pointing group 
		with the given rate

		Arguments:
		- rate: rate of motion in pixel/hour (numpy array or float)
		
		Returns:
		- detection probability, same shape as rate
		'''

		# r50_1, r50_2, kappa1, kappa2, c = params 
		return detprob_piecewise(rate, self.params, self.split)



	def observePopulation(self, population, lightcurve = None, keepall = False, expmin = 50, ccdmask = [2,61]):
		'''
		Uses the population's magnitudes and rates to check if a detection is observable given each pointing group's completeness

		Arguments:
		- population: Population object to be observed, requires magnitudes of each object
		- lightcurve: list (or single) of lightcurve objects to these can be applied for the objects. If a list is provided, make sure len(lightcurve) == len(population)
		- keepall: boolean, will split orbits between detections and non-detections if True
		- expmin (int): minimum number of exposures the source has to be inside the CCD
		- ccdmask (dictionary of list): dictionary with a  list of masked CCDs per PG
		'''
		try:
			population.observations
		except:
			raise AttributeError("Population has no observations")


		try:
			self.pointing_groups
		except:
			self.createPointingGroup()

		#exp = tb.join(self.exp, self.pg)

		#we don't need all exposures - should save memory
		#exp = self.exp[np.isin(self.exp['EXPNUM'], population.observations['EXPNUM'])]

		obs = tb.join(population.observations, self.exp['EXPNUM', 'BAND', 'MJD_MID'])
		obs.rename_column('MJD_MID', 'MJD')

		# remove what doesn't have observations before joining!
		mags = population.mag_obs[np.isin(population.mag_obs['ORBITID'], np.unique(obs['ORBITID']))]
		
		obs = tb.join(population.observations, mags)

		population.obs = obs

		if lightcurve is not None:
			population.generateLightCurve(lightcurve)

		del mags


		## now onto rates, the main change here 
		r = [] 
		for i in self.pg['POINTING_GROUP']:
			rate = self.pointing_groups[i].calculateRate(population, expmin, ccdmask, compute_mag = True)
			rate['POINTING_GROUP'] = i
			r.append(rate)

		r = tb.vstack(r)

		population.rates = r #tb.join(population.obs, r)
		population.rates = tb.join(population.rates, self.pg['POINTING_GROUP', 'EXPNUM1', 'EXPNUM2', 'MJD1', 'MJD2', 'm25', 'c', 'k1', 'k2'])

		### rate selection function

		denom  = (1. + np.exp(population.rates['k1'] * (population.rates['MAG'] - population.rates['m25'])))
		denom *= (1. + np.exp(population.rates['k2'] * (population.rates['MAG'] - population.rates['m25'])))
		population.rates['DETPROB'] = population.rates['c']/denom
														


		population.rates['RANDOM'] = np.random.rand(len(population.rates))
		del population.rates['m25', 'c', 'k1', 'k2']


		population.rates['DETPROB'] *= self.probRate(population.rates['RATE'])


		if not keepall:
			ndet = population.rates[population.rates['DETPROB'] < population.rates['RANDOM']]
			obs = population.rates[population.rates['DETPROB'] > population.rates['RANDOM']]
			population.ndet = ndet
		else:
			obs = population.rates


		p1 = obs['ORBITID', 'EXPNUM1', 'MJD1', 'POINTING_GROUP', 'MAG', 'RATE', 'DETPROB', 'RANDOM']
		p1.rename_column('EXPNUM1', 'EXPNUM')
		p1.rename_column('MJD1', 'MJD')

		p2 = obs['ORBITID', 'EXPNUM2', 'MJD2', 'POINTING_GROUP', 'MAG', 'RATE', 'DETPROB', 'RANDOM']
		p2.rename_column('EXPNUM2', 'EXPNUM')
		p2.rename_column('MJD2', 'MJD')
		p = tb.vstack([p1, p2])
		p.rename_column('MAG', 'MEAN_MAG')

		population.detections = tb.join(p, population.obs)
